import torch
import numpy as np
import ood_detectors.sde as sde_lib
import ood_detectors.ood_utils as ood_utils
import ood_detectors.ema as ema



def get_sde_loss_fn(
    sde, reduce_mean=True, continuous=True, likelihood_weighting=True, eps=1e-5
):
    """Create a loss function for training with arbirary SDEs.

    Args:
      sde: An `sde_lib.SDE` object that represents the forward SDE.
      train: `True` for training loss and `False` for evaluation loss.
      reduce_mean: If `True`, average the loss across data dimensions. Otherwise sum the loss across data dimensions.
      continuous: `True` indicates that the model is defined to take continuous time steps. Otherwise it requires
        ad-hoc interpolation to take continuous time steps.
      likelihood_weighting: If `True`, weight the mixture of score matching losses
        according to https://arxiv.org/abs/2101.09258.

    Returns:
      A loss function.
    """
    reduce_op = (
        torch.mean
        if reduce_mean
        else lambda *args, **kwargs: 0.5 * torch.sum(*args, **kwargs)
    )

    def loss_fn(model, batch):
        """Compute the loss function.

        Args:
          model: A score model.
          batch: A mini-batch of training data.

        Returns:
          loss: A scalar that represents the average loss value across the mini-batch.
        """
        score_fn = ood_utils.get_score_fn(sde, model, continuous=continuous)
        t = torch.rand(batch.shape[0], device=batch.device) * (sde.T - eps) + eps
        z = torch.randn_like(batch)
        mean, std = sde.marginal_prob(batch, t)
        perturbed_data = mean + std[:, None] * z
        score = score_fn(perturbed_data, t)

        if not likelihood_weighting:
            losses = torch.square(score * std[:, None] + z)
            losses = reduce_op(losses.reshape(losses.shape[0], -1), dim=-1)
        else:
            g2 = sde.sde(torch.zeros_like(batch), t)[1] ** 2
            losses = torch.square(score + z / std[:, None])
            losses = reduce_op(losses.reshape(losses.shape[0], -1), dim=-1) * g2

        loss = torch.mean(losses)
        return loss

    return loss_fn


def get_smld_loss_fn(vesde, reduce_mean=False):
    """Legacy code to reproduce previous results on SMLD(NCSN). Not recommended for new work."""
    assert isinstance(vesde, sde_lib.VESDE), "SMLD training only works for VESDEs."

    # Previous SMLD models assume descending sigmas
    smld_sigma_array = torch.flip(vesde.discrete_sigmas, dims=(0,))
    reduce_op = (
        torch.mean
        if reduce_mean
        else lambda *args, **kwargs: 0.5 * torch.sum(*args, **kwargs)
    )

    def loss_fn(model, batch):
        labels = torch.randint(0, vesde.N, (batch.shape[0],), device=batch.device)
        sigmas = smld_sigma_array.to(batch.device)[labels]
        noise = torch.randn_like(batch) * sigmas[:, None]
        perturbed_data = noise + batch
        score = model(perturbed_data, labels)
        target = -noise / (sigmas**2)[:, None]
        losses = torch.square(score - target)
        losses = reduce_op(losses.reshape(losses.shape[0], -1), dim=-1) * sigmas**2
        loss = torch.mean(losses)
        return loss

    return loss_fn


def get_ddpm_loss_fn(vpsde, reduce_mean=True):
    """Legacy code to reproduce previous results on DDPM. Not recommended for new work."""
    assert isinstance(vpsde, sde_lib.VPSDE), "DDPM training only works for VPSDEs."

    reduce_op = (
        torch.mean
        if reduce_mean
        else lambda *args, **kwargs: 0.5 * torch.sum(*args, **kwargs)
    )

    def loss_fn(model, batch):
        labels = torch.randint(0, vpsde.N, (batch.shape[0],), device=batch.device)
        sqrt_alphas_cumprod = vpsde.sqrt_alphas_cumprod.to(batch.device)
        sqrt_1m_alphas_cumprod = vpsde.sqrt_1m_alphas_cumprod.to(batch.device)
        noise = torch.randn_like(batch)
        perturbed_data = (
            sqrt_alphas_cumprod[labels, None] * batch
            + sqrt_1m_alphas_cumprod[labels, None] * noise
        )
        score = model(perturbed_data, labels)
        losses = torch.square(score - noise)
        losses = reduce_op(losses.reshape(losses.shape[0], -1), dim=-1)
        loss = torch.mean(losses)
        return loss

    return loss_fn


class SDE_EMA_Warmup_GradClip:
    def __init__(
        self,
        sde,
        model,
        optimizer,
        ema_rate=0.999,
        warmup=5000,
        grad_clip=1,
        reduce_mean=True,
        continuous=True,
        likelihood_weighting=False,
    ):
        """Create a one-step training/evaluation class.
        """
        if continuous:
            self.loss_fn = get_sde_loss_fn(
                sde,
                reduce_mean=reduce_mean,
                continuous=True,
                likelihood_weighting=likelihood_weighting,
            )
        else:
            assert (
                not likelihood_weighting
            ), "Likelihood weighting is not supported for original SMLD/DDPM training."
            if isinstance(sde, sde_lib.VESDE):
                self.loss_fn = get_smld_loss_fn(sde, reduce_mean=reduce_mean)
            elif isinstance(sde, sde_lib.VPSDE):
                self.loss_fn = get_ddpm_loss_fn(sde, reduce_mean=reduce_mean)
            else:
                raise ValueError(
                    f"Discrete training for {sde.__class__.__name__} is not recommended."
                )

        self.optimizer = optimizer
        self.ema = ema.ExponentialMovingAverage(model.parameters(), decay=ema_rate)
        self.step = 0
        self.warmup = warmup
        self.grad_clip = grad_clip
        self.lr = optimizer.param_groups[0]["lr"]
        # self.scaler = torch.cuda.amp.GradScaler()

    def __call__(self, model, x, train=True):
        """Running one step of training or evaluation.
        Returns:
        loss: The average loss value of this state.
        """

        if train:
            self.optimizer.zero_grad()
            loss = self.loss_fn(model, x)
            loss.backward()
            if self.warmup > 0:
                for g in self.optimizer.param_groups:
                    g["lr"] = self.lr * np.minimum(self.step / self.warmup, 1.0)
            if self.grad_clip >= 0:
                torch.nn.utils.clip_grad_norm_(
                    model.parameters(), max_norm=self.grad_clip
                )
            self.optimizer.step()
            self.step += 1
            if self.ema is not None:
                self.ema.update(model.parameters())
        else:
            with torch.no_grad():
                if self.ema is not None:
                    self.ema.store(model.parameters())
                    self.ema.copy_to(model.parameters())
                loss = self.loss_fn(model, x)
                if self.ema is not None:
                    self.ema.restore(model.parameters())
        return loss


class SDE_LRS_BF16:
    def __init__(
        self,
        sde,
        model,
        optimizer,
        total_steps=100000,
        reduce_mean=True,
        continuous=True,
        likelihood_weighting=False,
    ):
        """Create a one-step training/evaluation class.
        """
        if continuous:
            self.loss_fn = get_sde_loss_fn(
                sde,
                reduce_mean=reduce_mean,
                continuous=True,
                likelihood_weighting=likelihood_weighting,
            )
        else:
            assert (
                not likelihood_weighting
            ), "Likelihood weighting is not supported for original SMLD/DDPM training."
            if isinstance(sde, sde_lib.VESDE):
                self.loss_fn = get_smld_loss_fn(sde, reduce_mean=reduce_mean)
            elif isinstance(sde, sde_lib.VPSDE):
                self.loss_fn = get_ddpm_loss_fn(sde, reduce_mean=reduce_mean)
            else:
                raise ValueError(
                    f"Discrete training for {sde.__class__.__name__} is not recommended."
                )

        self.optimizer = optimizer
        self.lr = optimizer.param_groups[0]["lr"]
        self.scaler = torch.cuda.amp.GradScaler()
        self.total_steps = total_steps
        self.lrs = torch.optim.lr_scheduler.OneCycleLR(
            self.optimizer,
            max_lr=self.lr,
            total_steps=self.total_steps,
        )

    def __call__(self, model, x, *args, **kwargs):
        """Running one step of training or evaluation.
        Returns:
        loss: The average loss value of this state.
        """

        self.optimizer.zero_grad()
        with torch.cuda.amp.autocast(dtype=torch.bfloat16):
            loss = self.loss_fn(model, x)
        self.scaler.scale(loss).backward()
        self.scaler.step(self.optimizer)
        self.scaler.update()
        self.lrs.step()


        return loss
    
class SDE_BF16:
    def __init__(
        self,
        sde,
        model,
        optimizer,
        reduce_mean=True,
        continuous=True,
        likelihood_weighting=False,
    ):
        """Create a one-step training/evaluation class.
        """
        if continuous:
            self.loss_fn = get_sde_loss_fn(
                sde,
                reduce_mean=reduce_mean,
                continuous=True,
                likelihood_weighting=likelihood_weighting,
            )
        else:
            assert (
                not likelihood_weighting
            ), "Likelihood weighting is not supported for original SMLD/DDPM training."
            if isinstance(sde, sde_lib.VESDE):
                self.loss_fn = get_smld_loss_fn(sde, reduce_mean=reduce_mean)
            elif isinstance(sde, sde_lib.VPSDE):
                self.loss_fn = get_ddpm_loss_fn(sde, reduce_mean=reduce_mean)
            else:
                raise ValueError(
                    f"Discrete training for {sde.__class__.__name__} is not recommended."
                )

        self.optimizer = optimizer
        self.scaler = torch.cuda.amp.GradScaler()
      
    def __call__(self, model, x, *args, **kwargs):
        """Running one step of training or evaluation.
        Returns:
        loss: The average loss value of this state.
        """

        self.optimizer.zero_grad()
        with torch.cuda.amp.autocast(dtype=torch.bfloat16):
            loss = self.loss_fn(model, x)
        self.scaler.scale(loss).backward()
        self.scaler.step(self.optimizer)
        self.scaler.update()
        return loss


import torch
import ood_detectors.likelihood as likelihood
import numpy as np

class RDM():
    def __init__(self, feat_dim, number_of_RDMs=2):
        super().__init__()
        self.feat_dim = feat_dim
        self.number_of_RDMs = number_of_RDMs
        self.ood_detectors = [
            likelihood.subVPSDE_RDM(feat_dim) for _ in range(number_of_RDMs)
        ]
        self.name = f"RDM_{self.ood_detectors[0].name}x{number_of_RDMs}"
        self.device = "cpu"

    def to(self, device):
        for ood_detector in self.ood_detectors:
            ood_detector.to(device)
        self.device = device
        return self

    def load_state_dict(self, state_dict):
        for ood_detector, state_dict_ in zip(self.ood_detectors, state_dict):
            ood_detector.load_state_dict(state_dict_)

    def state_dict(self):
        return [ood_detector.state_dict() for ood_detector in self.ood_detectors]
    
    def fit(self, dataset, n_epochs, batch_size, num_workers=0, update_fn=None, verbose=True, collate_fn=None):
        losses = []
        for ood_detector in self.ood_detectors:
            loss = ood_detector.fit(dataset, n_epochs, batch_size, num_workers, update_fn, verbose, collate_fn)
            losses.append(loss)
        return losses
    
    def predict(self, x, *args, reduce=True, **kwargs):
        if reduce:
            return np.stack([ood_detector.predict(x,*args, **kwargs) for ood_detector in self.ood_detectors]).mean(axis=0)
        else:
            return np.stack([ood_detector.predict(x,*args, **kwargs) for ood_detector in self.ood_detectors])
    


        
    
        


import numpy as np
import torch
import tqdm

def train(dataset, model, update_fn, n_epochs, batch_size, device, num_workers=0, verbose=True, tw=None, collate_fn=None):
    """
    Trains a model on a given dataset for a specified number of epochs.

    Args:
        dataset (torch.utils.data.Dataset): The training dataset.
        model (torch.nn.Module): The model to be trained.
        update_fn (callable): The function that updates the model parameters.
        n_epochs (int): The number of training epochs.
        batch_size (int): The batch size for training.
        device (torch.device): The device to run the training on.
        num_workers (int, optional): The number of workers for data loading. Defaults to 0.
        verbose (bool, optional): Whether to display training progress. Defaults to True.
        tw (tensorboardX.SummaryWriter, optional): TensorboardX SummaryWriter for logging. Defaults to None.
        lrs (bool, optional): Whether to use learning rate scheduling. Defaults to True.

    Returns:
        list: A list of average epoch losses.
    """
    if verbose:
        print(f'Training for {n_epochs} epochs...')
    if collate_fn is None:
        data_loader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=True, num_workers=num_workers, drop_last=True)
    else:
        data_loader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=True, num_workers=num_workers, drop_last=True, collate_fn=collate_fn)
    if verbose:
        epochs = tqdm.trange(n_epochs)
    else:
        epochs = range(n_epochs)
    avg_epoch_loss = []
    model.train()
    for epoch in epochs:
        avg_loss = 0.
        num_items = 0
        epoch_loss = []
        for x in data_loader:
            x = x.to(device)
            loss = update_fn(model, x)
            avg_loss += loss.item() * x.shape[0]
            num_items += x.shape[0]
            epoch_loss.append(loss.item())
        epoch_loss = np.mean(epoch_loss)
        avg_epoch_loss.append(epoch_loss)
        if tw is not None:
            tw.add_scalar('Loss/train', epoch_loss, epoch)
        # Print the averaged training loss so far.
        if verbose:
            epochs.set_description('Average Loss: {:5f}'.format(avg_loss / num_items))
    return avg_epoch_loss


def inference(dataset, model, ode_likelihood, batch_size, device, num_workers=0, verbose=True, collate_fn=None):
    """
    Performs inference using a trained model on a given dataset.

    Args:
        dataset (torch.utils.data.Dataset): The dataset for inference.
        model (torch.nn.Module): The trained model.
        ode_likelihood (callable): The function that computes the likelihood.
        batch_size (int): The batch size for inference.
        device (torch.device): The device to run the inference on.
        num_workers (int, optional): The number of workers for data loading. Defaults to 0.
        verbose (bool, optional): Whether to display inference progress. Defaults to True.

    Returns:
        numpy.ndarray: An array of inference scores.
    """
    if collate_fn is None:
        data_loader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=True, num_workers=num_workers)
    else:
        data_loader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=True, num_workers=num_workers, collate_fn=collate_fn)
    all_bpds = 0.
    all_items = 0
    score_id = []
    if verbose:
        data_iter = tqdm.tqdm(data_loader)
    else:
        data_iter = data_loader
    model.eval()

    for x in data_iter:
        x = x.to(device)
        bpd = ode_likelihood(model=model, x=x)
        all_bpds += bpd.sum()
        all_items += bpd.shape[0]
        score_id.append(bpd.cpu().numpy())

        if verbose:
            data_iter.set_description("Average bits/dim: {:5f}".format(all_bpds / all_items))

    return np.concatenate(score_id)




import ood_detectors.vision as vision
import ood_detectors.ops_utils as ops_utils
import torch
import random
import pathlib
import pickle


import multiprocessing as mp

def encode_dataset(encoder_name, dataset_name, device):
    data_root = '/mnt/data/arty/data'
    image_list_root = '/mnt/data/arty/data/benchmark_imglist'
    features = pathlib.Path("/mnt/data/arty/data/features_opt_extra")
    features.mkdir(exist_ok=True)
    encoder_path = features / encoder_name
    encoder_path.mkdir(exist_ok=True)
    dataset_name_path = encoder_path / dataset_name
    dataset_name_path.mkdir(exist_ok=True)
    encoder = vision.get_encoder(encoder_name)
    encoder.eval()
    datasets = vision.get_datasets(dataset_name, data_root=data_root, image_list_root=image_list_root, transform=encoder.transform)
    for type_name, datasets in datasets.items():
        if type_name == 'csid':
            continue
        type_name_path = dataset_name_path / type_name
        type_name_path.mkdir(exist_ok=True)
        for name, dataset in datasets.items():
            name_path = type_name_path / name
            name_path.mkdir(exist_ok=True)
            print(encoder_name, dataset_name, name, len(dataset))
            features_path = name_path / 'feature_data.pkl'
            if features_path.exists():
                continue
            features_vectors = vision.extract_features(encoder, dataset, batch_size=64, num_workers=2, device=device)
            data = {
                'features': features_vectors,
                'encoder': encoder_name,
                'target_dataset': dataset_name,
                'dataset': name,
                'type': type_name
            }
            with open(features_path, 'wb') as f:
                pickle.dump(data, f)
def main():

    # encoders = ['repvgg', 'resnet50d', 'swin', 'deit', 'dino', 'dinov2', 'vit', 'clip', 'vit_b16', 'swin_t']
    encoders = ['vit_b16', 'swin_t']
    # datasets = ['imagenet', 'imagenet200', 'cifar10', 'cifar100']
    datasets = ['imagenet']
    gpu_nodes = [0, 1, 2, 3]

    jobs = []
    for encoder_name in encoders:
        for dataset_name in datasets:
            jobs.append((encoder_name, dataset_name))
        
    # for encoder_name, dataset_name in zip(['resnet18_32x32_cifar10_open_ood', 
    #                             'resnet18_32x32_cifar100_open_ood', 
    #                             'resnet18_224x224_imagenet200_open_ood', 
    #                             'resnet50_224x224_imagenet_open_ood'],
    #                             ['cifar10', 'cifar100', 'imagenet200', 'imagenet']):
    #     jobs.append((encoder_name, dataset_name))

    print(len(jobs))
    random.shuffle(jobs)


    ops_utils.parallelize(encode_dataset, jobs, gpu_nodes, verbose=True, timeout=60*60*24)

if __name__ == '__main__':
    mp.freeze_support()
    main()


import numpy as np

def multicol(name, space=2, bars=False):
    if bars:
        return r"\multicolumn{" +str(space)+r"}{|c|}{\textbf{" + name + r"}}"
    return r"\multicolumn{" +str(space)+r"}{c}{\textbf{" + name + r"}}"

def get_names(data, dataset):
    far_names = None
    near_names = None
    for d in data.values():
        if dataset not in d:
            continue
        far_names = []
        near_names = []

        for _dataset in d[dataset]["farood"]:
            if _dataset["dataset"] == "imagenet-o" and dataset != "imagenet_sub":
                continue
            far_names.append(_dataset["dataset"])
       
        for _dataset in d[dataset]["nearood"]:
            if _dataset["dataset"] == "imagenet-o" and dataset != "imagenet_sub":
                continue
            near_names.append(_dataset["dataset"])

        return sorted(far_names), sorted(near_names)
    return far_names, near_names
def create_latex_table(result, method, dataset, encoders):
    far_names, near_names = get_names(result[method], dataset)
    if far_names is None:
        return ""

    total_items = len(far_names) + len(near_names)

    header = r"\begin{table}[ht]"+ "\n"
    header += r"\caption{" +f'Result for '+ method.replace("_", "\\_") + ' on '+ dataset.replace("_", "\\_") + r"}"+ "\n"

    header += r"""\centering
\resizebox{\textwidth}{!}{% Resize table to fit within \textwidth horizontally
"""
    header += r"\begin{tabular}{@{}l*{" + str(total_items+1) + r"}{SS}@{}}" + "\n"
    header += r"\toprule" + "\n"

    description = r"\textbf{Encoder} & "  + " & ".join(multicol(name.replace("_", r"\_")) for name in near_names+far_names ) + " & " + multicol(" Average ") + r" \\" + "\n"
    description +=  r" & {\footnotesize AUROC} $\uparrow$ & {\footnotesize FPR95} $\downarrow$ "*(total_items +1)+ r" \\" + "\n"
    midrule = r"\midrule" + "\n"
    footer = r"\label{tab:" + f"{method}_{dataset}" + r"}" + "\n"
    footer += r"""
\bottomrule
\end{tabular}
}
\end{table}
"""+"\n"

    rows = [" & "   + multicol("Near OOD", len(near_names)*2, True) + " & " + multicol("Far OOD", len(far_names)*2, True) + r" & \\" + "\n"]#" & " + multicol("AVG",2, True) +" & " +multicol("AVG", 2, True) + r" & \\" + "\n"]
    # rows = []

    lookup = {}
    max_data = {}
    for encoder in sorted(encoders):
        if encoder not in result[method]:
            continue
        if dataset not in result[method][encoder]:
            continue
        for name, data in result[method][encoder][dataset].items():
            if name not in ["nearood", "farood"]:
                continue
            lookup_tmp = {d["dataset"]: d["metrics"] for d in data}
            if encoder not in lookup:
                lookup[encoder] = {}
            lookup[encoder].update(lookup_tmp)
 
    for e in encoders:
        if e not in lookup:
            continue
        for data_name in lookup[e]:
            if data_name not in lookup[e]:
                continue
            if data_name not in max_data:
                max_data[data_name] = {
                    "AUC": 0,
                    "FPR_95": 1
                }
            if lookup[e][data_name]["AUC"] > max_data[data_name]["AUC"]:
                max_data[data_name]["AUC"] = lookup[e][data_name]["AUC"]
            if lookup[e][data_name]["FPR_95"] < max_data[data_name]["FPR_95"]:
                max_data[data_name]["FPR_95"] = lookup[e][data_name]["FPR_95"]

    for encoder in sorted(encoders):
        if encoder not in result[method]:
            continue
        if dataset not in result[method][encoder]:
            continue
        if "resnet18_" in encoder:
            row = [r"resnet18\_open\_ood"]
        elif "resnet50_" in encoder:
            row = [r"resnet50\_open\_ood"]
        elif "swin_t" == encoder:
            row = [r"swin\_t\_open\_ood"]
        elif "vit_b16" == encoder:
            row = [r"vit\_b16\_open\_ood"]
        else:
            row = [encoder.replace("_", "\\_")]
        near_auc = 0
        near_fpr = 0
        count = 0
        for data_names in near_names:
            data_res = max_data[data_names]
            max_auc = data_res["AUC"]
            min_fpr = data_res["FPR_95"]
            if data_names in lookup[encoder]:
                metrics = lookup[encoder][data_names]
                if metrics["AUC"] == max_auc:
                    row.append(r"\textbf{" + f"{metrics['AUC']*100:.2f}" + r"}")
                else:
                    row.append(f"{metrics['AUC']*100:.2f}")
                near_auc += metrics["AUC"]
                if metrics["FPR_95"] == min_fpr:
                    row.append(r"\textbf{" + f"{metrics['FPR_95']*100:.2f}" + r"}")
                else:
                    row.append(f"{metrics['FPR_95']*100:.2f}")
                near_fpr += metrics["FPR_95"]
                count += 1
            else:
                row.append("-")
                row.append("-")
        far_auc = 0
        far_fpr = 0
        count2 = 0
        for data_names in far_names:
            data_res = max_data[data_names]
            max_auc = data_res["AUC"]
            min_fpr = data_res["FPR_95"]
            if data_names in lookup[encoder]:
                metrics = lookup[encoder][data_names]
                if metrics["AUC"] == max_auc:
                    row.append(r"\textbf{" + f"{metrics['AUC']*100:.2f}" + r"}")
                else:
                    row.append(f"{metrics['AUC']*100:.2f}")
                far_auc += metrics["AUC"]
                if metrics["FPR_95"] == min_fpr:
                    row.append(r"\textbf{" + f"{metrics['FPR_95']*100:.2f}" + r"}")
                else:
                    row.append(f"{metrics['FPR_95']*100:.2f}")
                far_fpr += metrics["FPR_95"]
                count2 += 1
            else:
                row.append("-")
                row.append("-")
        # row.append(f"{near_auc/count*100:.2f}")
        # row.append(f"{near_fpr/count*100:.2f}")
        # row.append(f"{(far_auc)/count2*100:.2f}")
        # row.append(f"{(far_fpr)/count2*100:.2f}")

        row.append(f"{(far_auc+near_auc)/(count+count2)*100:.2f}")
        row.append(f"{(far_fpr+near_fpr)/(count+count2)*100:.2f}")
        rows.append(" & ".join(row) + r" \\")

    return header + description + midrule + "\n".join(rows) + footer

def create_latex_table_mean(results, method, dataset, encoders):
    methods = ['VESDE', 'VPSDE', 'subVPSDE'] 
    experiments = ['results', 'results_v2', 'results_v3']
    result = results[experiments[0]]
    far_names, near_names = get_names(result[method], dataset)
    
    if far_names is None:
        return ""

    total_items = len(far_names) + len(near_names)

    header = r"\begin{table}[ht]"+ "\n"
    header += r"\caption{" +f'Result for '+ method.replace("_", "\\_") + ' on '+ dataset.replace("_", "\\_") + r"}"+ "\n"

    header += r"""\centering
\resizebox{\textwidth}{!}{% Resize table to fit within \textwidth horizontally
"""
    header += r"\begin{tabular}{@{}l*{" + str(total_items+1) + r"}{SS}@{}}" + "\n"
    header += r"\toprule" + "\n"

    description = r"\textbf{Encoder} & "  + " & ".join(multicol(name.replace("_", r"\_")) for name in near_names+far_names ) + " & " + multicol(" Average ") + r" \\" + "\n"
    description +=  r" & {\footnotesize AUROC} $\uparrow$ & {\footnotesize FPR95} $\downarrow$ "*(total_items +1)+ r" \\" + "\n"
    midrule = r"\midrule" + "\n"
    footer = r"\label{tab:" + f"{method}_{dataset}" + r"}" + "\n"
    footer += r"""
\bottomrule
\end{tabular}
}
\end{table}
"""+"\n"

    rows = [" & "   + multicol("Near OOD", len(near_names)*2, True) + " & " + multicol("Far OOD", len(far_names)*2, True) + r" & \\" + "\n"]#" & " + multicol("AVG",2, True) +" & " +multicol("AVG", 2, True) + r" & \\" + "\n"]
    # rows = []

    lookup = {}
    max_data = {}
    for _experiment in experiments:
        #print(_experiment)
        _result = results[_experiment]
        for _method in methods:
            for encoder in sorted(encoders):
                if _method not in _result:
                    continue
                if encoder not in _result[_method]:
                    continue
                if dataset not in _result[_method][encoder]:
                    continue
                for name, data in _result[_method][encoder][dataset].items():
                    if name not in ["nearood", "farood"]:
                        continue
                    if encoder not in lookup:
                        lookup[encoder] = {}
                    for d in data:
                        if d["dataset"] not in lookup[encoder]:
                            lookup[encoder][d["dataset"]] = {}
                        if "AUC" not in lookup[encoder][d["dataset"]]:
                            lookup[encoder][d["dataset"]]["AUC"] = []
                        lookup[encoder][d["dataset"]]["AUC"].append(d["metrics"]["AUC"])
                        if "FPR_95" not in lookup[encoder][d["dataset"]]:
                            lookup[encoder][d["dataset"]]["FPR_95"] = []
                        lookup[encoder][d["dataset"]]["FPR_95"].append(d["metrics"]["FPR_95"])

    # print(lookup)
    for e in encoders:
        if e not in lookup:
            continue
        for data_name in lookup[e]:
            if data_name not in lookup[e]:
                continue
            if data_name not in max_data:
                max_data[data_name] = {
                    "AUC": 0,
                    "FPR_95": 1
                }
            if np.mean(lookup[e][data_name]["AUC"]) > max_data[data_name]["AUC"]:
                max_data[data_name]["AUC"] = np.mean(lookup[e][data_name]["AUC"])
            if np.mean(lookup[e][data_name]["FPR_95"]) < max_data[data_name]["FPR_95"]:
                max_data[data_name]["FPR_95"] = np.mean(lookup[e][data_name]["FPR_95"])

    for encoder in sorted(encoders):
        if encoder not in result[method]:
            continue
        if dataset not in result[method][encoder]:
            continue
        if "resnet18_" in encoder:
            row = [r"resnet18\_open\_ood"]
        elif "resnet50_" in encoder:
            row = [r"resnet50\_open\_ood"]
        elif "swin_t" == encoder:
            row = [r"swin\_t\_open\_ood"]
        elif "vit_b16" == encoder:
            row = [r"vit\_b16\_open\_ood"]
        else:
            row = [encoder.replace("_", "\\_")]
        near_auc = 0
        near_fpr = 0
        count = 0
        for data_names in near_names+far_names:
            if data_names not in max_data:
                row.append("-")
                row.append("-")
                continue
            data_res = max_data[data_names]
            max_auc = data_res["AUC"]
            min_fpr = data_res["FPR_95"]
            if data_names in lookup[encoder]:
                metrics = lookup[encoder][data_names]
                auc_mean = np.mean(metrics['AUC'])
                auc_std = np.std(metrics['AUC'])
                if auc_mean == max_auc:
                    row.append(r"\textbf{" + f"{auc_mean*100:.2f}±{auc_std*100:.2f}" + r"}")
                else:
                    row.append(f"{auc_mean*100:.2f}±{auc_std*100:.2f}")
                near_auc += auc_mean
                fpr_mean = np.mean(metrics['FPR_95'])
                fpr_std = np.std(metrics['FPR_95'])
                if fpr_mean == min_fpr:
                    row.append(r"\textbf{" + f"{fpr_mean*100:.2f}±{fpr_std*100:.2f}" + r"}")
                else:
                    row.append(f"{fpr_mean*100:.2f}±{fpr_std*100:.2f}")
                near_fpr += fpr_mean
                count += 1
            else:
                row.append("-")
                row.append("-")
       
        # row.append(f"{near_auc/count*100:.2f}")
        # row.append(f"{near_fpr/count*100:.2f}")
        # row.append(f"{(far_auc)/count2*100:.2f}")
        # row.append(f"{(far_fpr)/count2*100:.2f}")
        if count == 0:
            row.append("-")
            row.append("-")
        else:
            row.append(f"{(near_auc)/(count)*100:.2f}")
            row.append(f"{(near_fpr)/(count)*100:.2f}")
        rows.append(" & ".join(row) + r" \\")

    return header + description + midrule + "\n".join(rows) + footer



def create_latex_table2(result, method, dataset, encoders):
    far_names, near_names = get_names(result[method], dataset)
    if far_names is None:
        return ""

    total_items = len(far_names) + len(near_names)

    header = r"\begin{table}[ht]"+ "\n"
    header += r"\caption{" +f'Result for ??? on '+ dataset.replace("_", "\\_") + r"}"+ "\n"

    header += r"""\centering
\resizebox{\textwidth}{!}{% Resize table to fit within \textwidth horizontally
"""
    header += r"\begin{tabular}{@{}l*{" + str(total_items+1) + r"}{SS}@{}}" + "\n"
    header += r"\toprule" + "\n"

    description = r"\textbf{Encoder} & "  + " & ".join(multicol(name.replace("_", r"\_")) for name in near_names+far_names ) + " & " + multicol(" Average ") + r" \\" + "\n"
    description +=  r" & {\footnotesize AUROC} $\uparrow$ & {\footnotesize FPR95} $\downarrow$ "*(total_items +1)+ r" \\" + "\n"
    midrule = r"\midrule" + "\n"
    footer = r"\label{tab:" + f"all" + r"}" + "\n"
    footer += r"""
\bottomrule
\end{tabular}
}
\end{table}
"""+"\n"

    # rows = [" & "   + multicol("Near OOD", len(near_names)*2, True) + " & " + multicol("Far OOD", len(far_names)*2, True) + r" & \\" + "\n"]#" & " + multicol("AVG",2, True) +" & " +multicol("AVG", 2, True) + r" & \\" + "\n"]
    rows = []

    lookup = {}
    max_data = {}
    for encoder in sorted(encoders):
        if encoder not in result[method]:
            continue
        if dataset not in result[method][encoder]:
            continue
        for name, data in result[method][encoder][dataset].items():
            if name not in ["nearood", "farood"]:
                continue
            lookup_tmp = {d["dataset"]: d["metrics"] for d in data}
            if encoder not in lookup:
                lookup[encoder] = {}
            lookup[encoder].update(lookup_tmp)
 
    for e in encoders:
        if e not in lookup:
            continue
        for data_name in lookup[e]:
            if data_name not in lookup[e]:
                continue
            if data_name not in max_data:
                max_data[data_name] = {
                    "AUC": 0,
                    "FPR_95": 1
                }
            if lookup[e][data_name]["AUC"] > max_data[data_name]["AUC"]:
                max_data[data_name]["AUC"] = lookup[e][data_name]["AUC"]
            if lookup[e][data_name]["FPR_95"] < max_data[data_name]["FPR_95"]:
                max_data[data_name]["FPR_95"] = lookup[e][data_name]["FPR_95"]

    for encoder in sorted(encoders):
        if encoder not in result[method]:
            continue
        if dataset not in result[method][encoder]:
            continue
        if "resnet18_" in encoder:
            row = [r"resnet18\_open\_ood"]
        elif "resnet50_" in encoder:
            row = [r"resnet50\_open\_ood"]
        elif "swin_t" == encoder:
            row = [r"swin\_t\_open\_ood"]
        elif "vit_b16" == encoder:
            row = [r"vit\_b16\_open\_ood"]
        row = [method.replace("_", "\\_")]
        near_auc = 0
        near_fpr = 0
        count = 0
        for data_names in near_names:
            data_res = max_data[data_names]
            max_auc = data_res["AUC"]
            min_fpr = data_res["FPR_95"]
            if data_names in lookup[encoder]:
                metrics = lookup[encoder][data_names]
                # if metrics["AUC"] == max_auc:
                #     row.append(r"\textbf{" + f"{metrics['AUC']*100:.2f}" + r"}")
                # else:
                row.append(f"{metrics['AUC']*100:.2f}")
                near_auc += metrics["AUC"]
                # if metrics["FPR_95"] == min_fpr:
                #     row.append(r"\textbf{" + f"{metrics['FPR_95']*100:.2f}" + r"}")
                # else:
                row.append(f"{metrics['FPR_95']*100:.2f}")
                near_fpr += metrics["FPR_95"]
                count += 1
            else:
                row.append("-")
                row.append("-")
        far_auc = 0
        far_fpr = 0
        count2 = 0
        for data_names in far_names:
            data_res = max_data[data_names]
            max_auc = data_res["AUC"]
            min_fpr = data_res["FPR_95"]
            if data_names in lookup[encoder]:
                metrics = lookup[encoder][data_names]
                # if metrics["AUC"] == max_auc:
                #     row.append(r"\textbf{" + f"{metrics['AUC']*100:.2f}" + r"}")
                # else:
                row.append(f"{metrics['AUC']*100:.2f}")
                far_auc += metrics["AUC"]
                # if metrics["FPR_95"] == min_fpr:
                #     row.append(r"\textbf{" + f"{metrics['FPR_95']*100:.2f}" + r"}")
                # else:
                row.append(f"{metrics['FPR_95']*100:.2f}")
                far_fpr += metrics["FPR_95"]
                count2 += 1
            else:
                row.append("-")
                row.append("-")
        # row.append(f"{near_auc/count*100:.2f}")
        # row.append(f"{near_fpr/count*100:.2f}")
        # row.append(f"{(far_auc)/count2*100:.2f}")
        # row.append(f"{(far_fpr)/count2*100:.2f}")

        row.append(f"{(far_auc+near_auc)/(count+count2)*100:.2f}")
        row.append(f"{(far_fpr+near_fpr)/(count+count2)*100:.2f}")
        rows.append(" & ".join(row) + r" \\")

    return header + description + midrule , "\n".join(rows) , footer

def create_latex_table3(result, method, dataset, encoders):
    far_names, near_names = get_names(result[method], dataset)
    if far_names is None:
        return ""

    total_items = len(far_names) + len(near_names)

    header = r"\begin{table}[ht]"+ "\n"
    header += r"\caption{" +f'Result for ??? on '+ dataset.replace("_", "\\_") + r"}"+ "\n"

    header += r"""\centering
\resizebox{\textwidth}{!}{% Resize table to fit within \textwidth horizontally
"""
    header += r"\begin{tabular}{@{}l*{" + str(total_items+1) + r"}{SS}@{}}" + "\n"
    header += r"\toprule" + "\n"

    description = r"\textbf{Encoder} & "  + " & ".join(multicol(name.replace("_", r"\_")) for name in near_names+far_names ) + " & " + multicol(" Average ") + r" \\" + "\n"
    description +=  r" & {\footnotesize AUROC} $\uparrow$ & {\footnotesize FPR95} $\downarrow$ "*(total_items +1)+ r" \\" + "\n"
    midrule = r"\midrule" + "\n"
    footer = r"\label{tab:" + f"all" + r"}" + "\n"
    footer += r"""
\bottomrule
\end{tabular}
}
\end{table}
"""+"\n"

    # rows = [" & "   + multicol("Near OOD", len(near_names)*2, True) + " & " + multicol("Far OOD", len(far_names)*2, True) + r" & \\" + "\n"]#" & " + multicol("AVG",2, True) +" & " +multicol("AVG", 2, True) + r" & \\" + "\n"]
    rows = []

    lookup = {}
    max_data = {}
    for encoder in sorted(encoders):
        if encoder not in result[method]:
            continue
        if dataset not in result[method][encoder]:
            continue
        for name, data in result[method][encoder][dataset].items():
            if name not in ["nearood", "farood"]:
                continue
            lookup_tmp = {d["dataset"]: d["metrics"] for d in data}
            if encoder not in lookup:
                lookup[encoder] = {}
            lookup[encoder].update(lookup_tmp)
 
    for e in encoders:
        if e not in lookup:
            continue
        for data_name in lookup[e]:
            if data_name not in lookup[e]:
                continue
            if data_name not in max_data:
                max_data[data_name] = {
                    "AUC": 0,
                    "FPR_95": 1
                }
            if lookup[e][data_name]["AUC"] > max_data[data_name]["AUC"]:
                max_data[data_name]["AUC"] = lookup[e][data_name]["AUC"]
            if lookup[e][data_name]["FPR_95"] < max_data[data_name]["FPR_95"]:
                max_data[data_name]["FPR_95"] = lookup[e][data_name]["FPR_95"]

    for encoder in sorted(encoders):
        if encoder not in result[method]:
            continue
        if dataset not in result[method][encoder]:
            continue
        if "resnet18_" in encoder:
            row = [r"resnet18\_open\_ood"]
        elif "resnet50_" in encoder:
            row = [r"resnet50\_open\_ood"]
        elif "swin_t" == encoder:
            row = [r"swin\_t\_open\_ood"]
        elif "vit_b16" == encoder:
            row = [r"vit\_b16\_open\_ood"]
        row = [method.replace("_", "\\_")+" "+encoder.replace("_", "\\_")]
        near_auc = 0
        near_fpr = 0
        count = 0
        for data_names in near_names:
            data_res = max_data[data_names]
            max_auc = data_res["AUC"]
            min_fpr = data_res["FPR_95"]
            if data_names in lookup[encoder]:
                metrics = lookup[encoder][data_names]
                # if metrics["AUC"] == max_auc:
                #     row.append(r"\textbf{" + f"{metrics['AUC']*100:.2f}" + r"}")
                # else:
                row.append(f"{metrics['AUC']*100:.2f}")
                near_auc += metrics["AUC"]
                # if metrics["FPR_95"] == min_fpr:
                #     row.append(r"\textbf{" + f"{metrics['FPR_95']*100:.2f}" + r"}")
                # else:
                row.append(f"{metrics['FPR_95']*100:.2f}")
                near_fpr += metrics["FPR_95"]
                count += 1
            else:
                row.append("-")
                row.append("-")
        far_auc = 0
        far_fpr = 0
        count2 = 0
        for data_names in far_names:
            data_res = max_data[data_names]
            max_auc = data_res["AUC"]
            min_fpr = data_res["FPR_95"]
            if data_names in lookup[encoder]:
                metrics = lookup[encoder][data_names]
                # if metrics["AUC"] == max_auc:
                #     row.append(r"\textbf{" + f"{metrics['AUC']*100:.2f}" + r"}")
                # else:
                row.append(f"{metrics['AUC']*100:.2f}")
                far_auc += metrics["AUC"]
                # if metrics["FPR_95"] == min_fpr:
                #     row.append(r"\textbf{" + f"{metrics['FPR_95']*100:.2f}" + r"}")
                # else:
                row.append(f"{metrics['FPR_95']*100:.2f}")
                far_fpr += metrics["FPR_95"]
                count2 += 1
            else:
                row.append("-")
                row.append("-")
        # row.append(f"{near_auc/count*100:.2f}")
        # row.append(f"{near_fpr/count*100:.2f}")
        # row.append(f"{(far_auc)/count2*100:.2f}")
        # row.append(f"{(far_fpr)/count2*100:.2f}")

        row.append(f"{(far_auc+near_auc)/(count+count2)*100:.2f}")
        row.append(f"{(far_fpr+near_fpr)/(count+count2)*100:.2f}")
        rows.append(" & ".join(row) + r" \\")

    return header + description + midrule , "\n".join(rows) , footer


import torch
from PIL import Image
from pathlib import Path
import os


def default_loader(path):
    return Image.open(path).convert('RGB')


def default_flist_reader(root, flist):
    """flist format: impath label\nimpath label\n."""
    imlist = []
    if flist == "all":
        for path in Path(root).rglob('*'):
            if path.is_file() and path.suffix.lower() in ['.jpg', '.png', '.jpeg']:
                imlist.append((str(path), 0))
    else:
        with open(flist, 'r') as rf:
            for line in rf.readlines():
                data = line.strip().rsplit(maxsplit=1)
                if len(data) == 2:
                    impath, imlabel = data
                else:
                    impath, imlabel = data[0], 0
                path = os.path.join(root, impath)
                # if not os.path.isfile(path):
                #     raise FileNotFoundError(f"Image file not found: {path}")
                imlist.append((str(path), int(imlabel)))
    return imlist


class ImageFilelist(torch.utils.data.Dataset):

    def __init__(self,
                 root,
                 flist,
                 transform=None,
                 target_transform=None,
                 flist_reader=default_flist_reader,
                 loader=default_loader):
        self.root = root
        self.imlist = flist_reader(root, flist)
        self.transform = transform
        self.target_transform = target_transform
        self.loader = loader

    def __getitem__(self, index):
        impath, target = self.imlist[index]
        img = self.loader(impath)
        if self.transform is not None:
            img = self.transform(img)
        if self.target_transform is not None:
            target = self.target_transform(target)

        return img, target

    def __len__(self):
        return len(self.imlist)



import numpy as np
from sklearn.covariance import EmpiricalCovariance
import torch


class Residual:
    """
    Residual class for outlier detection.

    Args:
        dims (int): Number of dimensions to consider for outlier detection. Default is 512.
        u (int): Mean value for data centering. Default is 0.

    Attributes:
        dims (int): Number of dimensions to consider for outlier detection.
        u (int): Mean value for data centering.
        name (str): Name of the Residual instance.

    Methods:
        fit(data, *args, **kwargs): Fit the Residual model to the given data.
        predict(data, *args, **kwargs): Predict the outlier scores for the given data.
        to(device): Move the Residual model to the specified device.
        state_dict(): Get the state dictionary of the Residual model.
        load_state_dict(state_dict): Load the state dictionary into the Residual model.

    """

    def __init__(self, dims=512, u=0):
        self.dims = dims
        self.u = u
        self.name = "Residual"

    def fit(self, data, *args, collate_fn=None, **kwargs):
        """
        Fit the Residual model to the given data.

        Args:
            data (array-like or torch.Tensor or torch.utils.data.DataLoader): Input data for fitting the model.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            list: An empty list.

        """
            
        if isinstance(data, (list, tuple)):
            data = np.array(data)
        elif isinstance(data, torch.Tensor):
            data = data.cpu().numpy()
        elif isinstance(data, torch.utils.data.Dataset):
            if collate_fn is None and getattr(data, 'collate_fn', None) is not None:
                collate_fn = data.collate_fn
            if collate_fn is None:
                data = np.vstack([x.cpu().numpy() for x, *_ in data])
            else:
                data = np.vstack([collate_fn([d]).cpu().numpy() for d in data])
        feat_dim = data.shape[-1]
        if self.dims is None:
            self.dims = 1000 if feat_dim >= 2048 else 512 
        if self.dims <= 1:
            self.dims = int(feat_dim * self.dims)
        if self.dims < 2:
            self.dims = 2
    
        ec = EmpiricalCovariance(assume_centered=True)
        ec.fit(data - self.u)
        eig_vals, eigen_vectors = np.linalg.eig(ec.covariance_)
        self.ns = np.ascontiguousarray(
            (eigen_vectors.T[np.argsort(eig_vals * -1)[self.dims :]]).T
        ).astype(np.float32)
        return [-1]

    def predict(self, data, batch_size=32, *args, collate_fn=None, **kwargs):
        """
        Predict the outlier scores for the given data in batches.

        Args:
            data (array-like, torch.Tensor, torch.utils.data.DataLoader, or torch.utils.data.Dataset): Input data for predicting the outlier scores.
            batch_size (int): The size of the batches to use for prediction.
            *args: Additional positional arguments.
            **kwargs: Additional keyword arguments.

        Returns:
            numpy.ndarray: Outlier scores for the input data.
        """
        def to_numpy(data_batch):
            if isinstance(data_batch, torch.Tensor):
                return data_batch.cpu().numpy()
            elif isinstance(data_batch, (list, tuple)):
                return np.array(data_batch)
            return data_batch

        if isinstance(data, (list, tuple, np.ndarray)):
            data = torch.tensor(data, dtype=torch.float32)
            dataset = torch.utils.data.TensorDataset(data)
            data_loader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=False)
        elif isinstance(data, torch.Tensor):
            dataset = torch.utils.data.TensorDataset(data)
            data_loader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=False)
        elif isinstance(data, torch.utils.data.Dataset):
            data_loader = torch.utils.data.DataLoader(data, batch_size=batch_size, shuffle=False, collate_fn=collate_fn)
        elif isinstance(data, torch.utils.data.DataLoader):
            data_loader = data
        else:
            raise TypeError("Unsupported data type: {}".format(type(data)))

        scores = []
        for batch in data_loader:
            batch = batch[0]  # TensorDataset returns a tuple
            batch_numpy = to_numpy(batch)
            batch_numpy = batch_numpy.astype(np.float32)
            batch_scores = np.linalg.norm((batch_numpy - self.u) @ self.ns, axis=-1)
            scores.append(batch_scores)

        return np.concatenate(scores)

    def __call__(self, *args, **kwargs):
        return self.predict(*args, **kwargs)

    def to(self, device):
        """
        Move the Residual model to the specified device.

        Args:
            device: Device to move the model to.

        """
        pass

    def state_dict(self):
        """
        Get the state dictionary of the Residual model.

        Returns:
            dict: State dictionary of the Residual model.

        """
        return {"dims": self.dims, "u": self.u, "ns": self.ns}

    def load_state_dict(self, state_dict):
        """
        Load the state dictionary into the Residual model.

        Args:
            state_dict (dict): State dictionary to load into the Residual model.

        Returns:
            self: Loaded Residual model.

        """
        self.dims = state_dict["dims"]
        self.u = state_dict["u"]
        self.ns = state_dict["ns"]
        return self



"""Abstract SDE classes, Reverse SDE, and VE/VP SDEs."""

import abc
import torch
import numpy as np

class SDE(abc.ABC):
    """SDE abstract class. Functions are designed for a mini-batch of inputs."""

    def __init__(self, N):
        """Construct an SDE.

        Args:
          N: number of discretization time steps.
        """
        super().__init__()
        self.N = N

    @property
    @abc.abstractmethod
    def T(self):
        """End time of the SDE."""
        pass

    @abc.abstractmethod
    def sde(self, x, t):
        pass

    @abc.abstractmethod
    def marginal_prob(self, x, t):
        """Parameters to determine the marginal distribution of the SDE, $p_t(x)$."""
        pass

    @abc.abstractmethod
    def prior_sampling(self, shape):
        """Generate one sample from the prior distribution, $p_T(x)$."""
        pass

    @abc.abstractmethod
    def prior_logp(self, z):
        """Compute log-density of the prior distribution.

        Useful for computing the log-likelihood via probability flow ODE.

        Args:
          z: latent code
        Returns:
          log probability density
        """
        pass

    def discretize(self, x, t):
        """Discretize the SDE in the form: x_{i+1} = x_i + f_i(x_i) + G_i z_i.

        Useful for reverse diffusion sampling and probabiliy flow sampling.
        Defaults to Euler-Maruyama discretization.

        Args:
          x: a torch tensor
          t: a torch float representing the time step (from 0 to `self.T`)

        Returns:
          f, G
        """
        dt = 1 / self.N
        drift, diffusion = self.sde(x, t)
        f = drift * dt
        G = diffusion * torch.sqrt(torch.tensor(dt, device=t.device))
        return f, G

    def reverse(self, score_fn, probability_flow=False):
        """Create the reverse-time SDE/ODE.

        Args:
          score_fn: A time-dependent score-based model that takes x and t and returns the score.
          probability_flow: If `True`, create the reverse-time ODE used for probability flow sampling.
        """
        N = self.N
        T = self.T
        sde_fn = self.sde
        discretize_fn = self.discretize

        # Build the class for reverse-time SDE.
        class RSDE(self.__class__):
            def __init__(self):
                self.N = N
                self.probability_flow = probability_flow

            @property
            def T(self):
                return T

            def sde(self, x, t):
                """Create the drift and diffusion functions for the reverse SDE/ODE."""
                drift, diffusion = sde_fn(x, t)
                score = score_fn(x, t)
                drift = drift - diffusion[:, None] ** 2 * score * (
                    0.5 if self.probability_flow else 1.0
                )
                # Set the diffusion function to zero for ODEs.
                diffusion = 0.0 if self.probability_flow else diffusion
                return drift, diffusion

            def discretize(self, x, t):
                """Create discretized iteration rules for the reverse diffusion sampler."""
                f, G = discretize_fn(x, t)
                rev_f = f - G[:, None] ** 2 * score_fn(x, t) * (
                    0.5 if self.probability_flow else 1.0
                )
                rev_G = torch.zeros_like(G) if self.probability_flow else G
                return rev_f, rev_G

        return RSDE()


class VPSDE(SDE):
    def __init__(self, beta_min=0.1, beta_max=20, N=1000):
        """Construct a Variance Preserving SDE.

        Args:
          beta_min: value of beta(0)
          beta_max: value of beta(1)
          N: number of discretization steps
        """
        super().__init__(N)
        self.beta_0 = beta_min
        self.beta_1 = beta_max
        self.N = N
        self.discrete_betas = torch.linspace(beta_min / N, beta_max / N, N)
        self.alphas = 1.0 - self.discrete_betas
        self.alphas_cumprod = torch.cumprod(self.alphas, dim=0)
        self.sqrt_alphas_cumprod = torch.sqrt(self.alphas_cumprod)
        self.sqrt_1m_alphas_cumprod = torch.sqrt(1.0 - self.alphas_cumprod)

    @property
    def T(self):
        return 1

    def sde(self, x, t):
        beta_t = self.beta_0 + t * (self.beta_1 - self.beta_0)
        drift = -0.5 * beta_t[:, None] * x
        diffusion = torch.sqrt(beta_t)
        return drift, diffusion

    def marginal_prob(self, x, t):
        log_mean_coeff = (
            -0.25 * t**2 * (self.beta_1 - self.beta_0) - 0.5 * t * self.beta_0
        )
        mean = torch.exp(log_mean_coeff[:, None]) * x
        std = torch.sqrt(1.0 - torch.exp(2.0 * log_mean_coeff))
        return mean, std

    def prior_sampling(self, shape):
        return torch.randn(*shape)

    def prior_logp(self, z):
        shape = z.shape
        N = np.prod(shape[1:])
        logps = -N / 2.0 * np.log(2 * np.pi) - torch.sum(z**2, dim=1) / 2.0
        return logps

    def discretize(self, x, t):
        """DDPM discretization."""
        timestep = (t * (self.N - 1) / self.T).long()
        beta = self.discrete_betas.to(x.device)[timestep]
        alpha = self.alphas.to(x.device)[timestep]
        sqrt_beta = torch.sqrt(beta)
        f = torch.sqrt(alpha)[:, None] * x - x
        G = sqrt_beta
        return f, G


class subVPSDE(SDE):
    def __init__(self, beta_min=0.1, beta_max=20, N=1000):
        """Construct the sub-VP SDE that excels at likelihoods.

        Args:
          beta_min: value of beta(0)
          beta_max: value of beta(1)
          N: number of discretization steps
        """
        super().__init__(N)
        self.beta_0 = beta_min
        self.beta_1 = beta_max
        self.N = N

    @property
    def T(self):
        return 1

    def sde(self, x, t):
        beta_t = self.beta_0 + t * (self.beta_1 - self.beta_0)
        drift = -0.5 * beta_t[:, None] * x
        discount = 1.0 - torch.exp(
            -2 * self.beta_0 * t - (self.beta_1 - self.beta_0) * t**2
        )
        diffusion = torch.sqrt(beta_t * discount)
        return drift, diffusion

    def marginal_prob(self, x, t):
        log_mean_coeff = (
            -0.25 * t**2 * (self.beta_1 - self.beta_0) - 0.5 * t * self.beta_0
        )
        mean = torch.exp(log_mean_coeff)[:, None] * x
        std = 1 - torch.exp(2.0 * log_mean_coeff)
        return mean, std

    def prior_sampling(self, shape):
        return torch.randn(*shape)

    def prior_logp(self, z):
        shape = z.shape
        N = np.prod(shape[1:])
        return -N / 2.0 * np.log(2 * np.pi) - torch.sum(z**2, dim=1) / 2.0


class VESDE(SDE):
    def __init__(self, sigma_min=0.01, sigma_max=50, N=1000):
        """Construct a Variance Exploding SDE.

        Args:
          sigma_min: smallest sigma.
          sigma_max: largest sigma.
          N: number of discretization steps
        """
        super().__init__(N)
        self.sigma_min = sigma_min
        self.sigma_max = sigma_max
        self.discrete_sigmas = torch.exp(
            torch.linspace(np.log(self.sigma_min), np.log(self.sigma_max), N)
        )
        self.N = N

    @property
    def T(self):
        return 1

    def sde(self, x, t):
        sigma = self.sigma_min * (self.sigma_max / self.sigma_min) ** t
        drift = torch.zeros_like(x)
        diffusion = sigma * torch.sqrt(
            torch.tensor(
                2 * (np.log(self.sigma_max) - np.log(self.sigma_min)), device=t.device
            )
        )
        return drift, diffusion

    def marginal_prob(self, x, t):
        std = self.sigma_min * (self.sigma_max / self.sigma_min) ** t
        mean = x
        return mean, std

    def prior_sampling(self, shape):
        return torch.randn(*shape) * self.sigma_max

    def prior_logp(self, z):
        shape = z.shape
        N = np.prod(shape[1:])
        return -N / 2.0 * np.log(2 * np.pi * self.sigma_max**2) - torch.sum(
            z**2, dim=1
        ) / (2 * self.sigma_max**2)

    def discretize(self, x, t):
        """SMLD(NCSN) discretization."""
        timestep = (t * (self.N - 1) / self.T).long()
        sigma = self.discrete_sigmas.to(t.device)[timestep]
        adjacent_sigma = torch.where(
            timestep == 0,
            torch.zeros_like(t),
            self.discrete_sigmas[timestep - 1].to(t.device),
        )
        f = torch.zeros_like(x)
        G = torch.sqrt(sigma**2 - adjacent_sigma**2)
        return f, G



import torch
import numpy as np
from torchdiffeq import odeint as odeint_torch
import ood_detectors.sde as sde_lib


def get_score_fn(sde, model, train=False, continuous=False):
    """Wraps `score_fn` so that the model output corresponds to a real time-dependent score function.

    Args:
      sde: An `sde_lib.SDE` object that represents the forward SDE.
      model: A score model.
      train: `True` for training and `False` for evaluation.
      continuous: If `True`, the score-based model is expected to directly take continuous time steps.

    Returns:
      A score function.
    """

    if isinstance(sde, sde_lib.VPSDE) or isinstance(sde, sde_lib.subVPSDE):

        def score_fn(x, t):
            # Scale neural network output by standard deviation and flip sign
            if continuous or isinstance(sde, sde_lib.subVPSDE):
                # For VP-trained models, t=0 corresponds to the lowest noise level
                # The maximum value of time embedding is assumed to 999 for
                # continuously-trained models.
                labels = t * 999
                score = model(x, labels)
                std = sde.marginal_prob(torch.zeros_like(x), t)[1]
            else:
                # For VP-trained models, t=0 corresponds to the lowest noise level
                labels = t * (sde.N - 1)
                score = model(x, labels)
                std = sde.sqrt_1m_alphas_cumprod.to(labels.device)[labels.long()]

            score = -score / std[:, None]
            return score

    elif isinstance(sde, sde_lib.VESDE):

        def score_fn(x, t):
            if continuous:
                labels = sde.marginal_prob(torch.zeros_like(x), t)[1]
            else:
                # For VE-trained models, t=0 corresponds to the highest noise level
                labels = sde.T - t
                labels *= sde.N - 1
                labels = torch.round(labels).long()

            score = model(x, labels)
            return score

    else:
        raise NotImplementedError(
            f"SDE class {sde.__class__.__name__} not yet supported."
        )

    return score_fn


def to_flattened_numpy(x):
    """Flatten a torch tensor `x` and convert it to numpy."""
    return x.detach().cpu().numpy().reshape((-1,))


def from_flattened_numpy(x, shape):
    """Form a torch tensor with the given `shape` from a flattened numpy array `x`."""
    return torch.from_numpy(x.reshape(shape))


def get_div_fn(fn):
    """Create the divergence function of `fn` using the Hutchinson-Skilling trace estimator."""

    def div_fn(x, t, eps):
        with torch.enable_grad():
            x.requires_grad_(True)
            fn_eps = torch.sum(fn(x, t) * eps)
            grad_fn_eps = torch.autograd.grad(fn_eps, x)[0]
        x.requires_grad_(False)
        return torch.sum(grad_fn_eps * eps, dim=tuple(range(1, len(x.shape))))

    return div_fn


def get_likelihood_fn(
    sde,
    hutchinson_type="Rademacher",
    rtol=1e-5,
    atol=1e-5,
    method="fehlberg2",
    eps=1e-5,
):
    """Create a function to compute the unbiased log-likelihood estimate of a given data point.
    # Solve the ODE
    # 'dopri8' 7s
    # 'dopri5' 1.9s - good same as scipy.solve_ivp rk45
    # 'bosh3' 2.5s
    # 'fehlberg2' 1.4s - is scipy.solve_ivp rkf45
    # 'adaptive_heun' 4s
    # 'euler' nan
    # 'midpoint' nan
    # 'rk4' 1s inaccurate
    # 'explicit_adams' 1s inaccurate
    # 'implicit_adams' 1s inaccurate
    # 'fixed_adams' 1s inaccurate
    # 'scipy_solver'

    Args:
      sde: A `sde_lib.SDE` object that represents the forward SDE.
      inverse_scaler: The inverse data normalizer.
      hutchinson_type: "Rademacher" or "Gaussian". The type of noise for Hutchinson-Skilling trace estimator.
      rtol: A `float` number. The relative tolerance level of the black-box ODE solver.
      atol: A `float` number. The absolute tolerance level of the black-box ODE solver.
      method: A `str`. The algorithm for the black-box ODE solver.
        See documentation for `scipy.integrate.solve_ivp`.
      eps: A `float` number. The probability flow ODE is integrated to `eps` for numerical stability.

    Returns:
      A function that a batch of data points and returns the log-likelihoods in bits/dim,
        the latent code, and the number of function evaluations cost by computation.
    """

    def drift_fn(model, x, t):
        """The drift function of the reverse-time SDE."""
        score_fn = get_score_fn(sde, model, continuous=True)
        # Probability flow ODE is a special case of Reverse SDE
        rsde = sde.reverse(score_fn, probability_flow=True)
        return rsde.sde(x, t)[0]

    def div_fn(model, x, t, noise):
        return get_div_fn(lambda xx, tt: drift_fn(model, xx, tt))(x, t, noise)

    def likelihood_fn(model, x):
        """Compute an unbiased estimate to the log-likelihood in bits/dim.

        Args:
          model: A score model.
          x: A PyTorch tensor.

        Returns:
          bpd: A PyTorch tensor of shape [batch size]. The log-likelihoods on `data` in bits/dim.
          z: A PyTorch tensor of the same shape as `data`. The latent representation of `data` under the
            probability flow ODE.
          nfe: An integer. The number of function evaluations used for running the black-box ODE solver.
        """
        device = x.device
        with torch.no_grad():
            shape = x.shape
            if hutchinson_type == "Gaussian":
                epsilon = torch.randn_like(x)
            elif hutchinson_type == "Rademacher":
                epsilon = torch.randint_like(x, low=0, high=2).float() * 2 - 1.0
            else:
                raise NotImplementedError(f"Hutchinson type {hutchinson_type} unknown.")

            def ode_func(t, data):
                sample = data[: shape.numel()].clone().reshape(shape).float()

                vec_t = torch.ones(sample.shape[0], device=sample.device) * t
                drift = drift_fn(model, sample, vec_t)
                div = div_fn(model, sample, vec_t, epsilon)
                return torch.cat([drift.reshape(-1), div], 0)

            init_state = torch.cat(
                [x.reshape(-1), torch.zeros(shape[0], device=device)], 0
            )
            timesteps = torch.tensor([eps, sde.T], device=device)

            # Solving the ODE
            res = odeint_torch(
                ode_func, init_state, timesteps, rtol=rtol, atol=atol, method=method
            )
            zp = res[-1]

            z = zp[: -shape[0]].reshape(shape)
            prior_logp = sde.prior_logp(z)

            delta_logp = zp[-shape[0] :].reshape(shape[0])

            bpd = -(prior_logp + delta_logp) / np.log(2)
            N = torch.prod(torch.tensor(shape[1:], device=device)).item()
            bpd = bpd / N + 8  # Convert log-likelihood to bits/dim
            return bpd

    return likelihood_fn



import torch
import pickle
import ood_detectors.sde as sde_lib
import ood_detectors.models as models
import ood_detectors.losses as losses
import ood_detectors.likelihood as likelihood
import ood_detectors.eval_utils as eval_utils
import ood_detectors.ops_utils as ops_utils
import functools
from ood_detectors.residual import Residual
import pathlib
import tqdm
import yaml
import multiprocessing as mp
import random

def run(conf, data, encoder, dataset, method, device, reduce_data_eval=-1, reduce_data_train=-1, verbose=False, checkpoints="results"):
    train_data_path = data[encoder][dataset]["id"]["train"]
    batch_size = conf.get('batch_size', 1024)
    with open(train_data_path, 'rb') as f:
        train_blob = pickle.load(f)
    data_train = train_blob['features']
    if reduce_data_train > 0:
        prem = torch.randperm(data_train.shape[0])
        data_train = data_train[prem[:reduce_data_train]]

    if method == 'Residual':
        dims = conf['dims']
        ood_model = Residual(dims=dims)
        if pathlib.Path(f"{checkpoints}/{method}_{dataset}_{encoder}.pt").exists():
            model_path = f"{checkpoints}/{method}_{dataset}_{encoder}.pt"
            checkpoint = torch.load(model_path)
            ood_model.load_state_dict(checkpoint['model'])
            loss = [checkpoint['results']['id']['loss']]
        else:
            loss = ood_model.fit(data_train)
    else:

        # Hyperparameters
        feat_dim = data_train.shape[-1]
        n_epochs = conf.get('n_epochs', 200)
        bottleneck_channels = conf['bottleneck_channels']
        num_res_blocks = conf['num_res_blocks']
        time_embed_dim = conf['time_embed_dim']
        dropout = conf['dropout']
        lr = conf['lr']
        beta1 = conf.get('beta1', 0.9)
        beta2 = conf.get('beta2', 0.999)
        eps = conf.get('eps', 1e-8)
        weight_decay = conf.get('weight_decay', 0)
        continuous = conf['continuous']
        reduce_mean = conf['reduce_mean']
        likelihood_weighting = conf['likelihood_weighting']

        if method == 'VESDE':
            sigma_min = conf['sigma_min']
            sigma_max = conf['sigma_max']
            sde = sde_lib.VESDE(sigma_min=sigma_min, sigma_max=sigma_max)
        elif method == 'VPSDE':
            beta_min = conf['beta_min']
            beta_max = conf['beta_max']
            sde = sde_lib.VPSDE(beta_min=beta_min, beta_max=beta_max)
        elif method == 'subVPSDE':
            beta_min = conf['beta_min']
            beta_max = conf['beta_max']
            sde = sde_lib.subVPSDE(beta_min=beta_min, beta_max=beta_max)

        model = models.SimpleMLP(
            channels=feat_dim,
            bottleneck_channels=bottleneck_channels,
            num_res_blocks=num_res_blocks,
            time_embed_dim=time_embed_dim,
            dropout=dropout,
        )

        optimizer = functools.partial(
                        torch.optim.Adam,
                        lr=lr,
                        betas=(beta1, beta2),
                        eps=eps,
                        weight_decay=weight_decay,
                        )

        ood_model = likelihood.Likelihood(
            sde = sde,
            model = model,
            optimizer = optimizer,
            ).to(device)

        update_fn = functools.partial(
            losses.SDE_BF16, 
            continuous=continuous,
            reduce_mean=reduce_mean,
            likelihood_weighting=likelihood_weighting,
            )
        if pathlib.Path(f"{checkpoints}/{method}_{dataset}_{encoder}.pt").exists():
            model_path = f"{checkpoints}/{method}_{dataset}_{encoder}.pt"
            checkpoint = torch.load(model_path)
            ood_model.load_state_dict(checkpoint['model'])
            loss = [checkpoint['results']['id']['loss']]
        else:
            loss = ood_model.fit(
                data_train,
                n_epochs=n_epochs,
                batch_size=batch_size,
                update_fn=update_fn,
                verbose=verbose,
            )

    score_id = ood_model.predict(data_train, batch_size, verbose=verbose)
    test_data_path = data[encoder][dataset]["id"]["test"]
    with open(test_data_path, 'rb') as f:
        test_blob = pickle.load(f)
    data_test = test_blob['features']
    if reduce_data_eval > 0:
        prem = torch.randperm(data_test.shape[0])
        data_test = data_test[prem[:reduce_data_eval]]

    score_ref = ood_model.predict(data_test, batch_size, verbose=verbose)
    results = {}
    scores = {
        'score_id': score_id,
        'score_ref': score_ref,
    }
    id_auc = eval_utils.auc(-score_ref, -score_id)
    id_fpr_95 = eval_utils.fpr(-score_ref, -score_id, 0.95)
    results['id'] = {'AUC': id_auc, 
                     'FPR_95': id_fpr_95, 
                     'score_id': score_id.mean().item(),
                     'score_ref': score_ref.mean().item(), 
                     'loss': loss[-1]}
    ood_auc_mean = []
    for name, datasets in data[encoder][dataset].items():
        if name == 'id':
            continue
        if name not in results:
            results[name] = {}
        for type_name, data in datasets.items():
            with open(data, 'rb') as f:
                ood_data = pickle.load(f)
            data = ood_data['features']
            if reduce_data_eval > 0:
                prem = torch.randperm(data.shape[0])
                data = data[prem[:reduce_data_eval]]
            score_ood = ood_model.predict(data, batch_size, verbose=verbose)
            ood_auc = eval_utils.auc(-score_ref, -score_ood)
            ood_fpr_95 = eval_utils.fpr(-score_ref, -score_ood, 0.95)
            results[name][type_name] = {'AUC': ood_auc, 
                                        'FPR_95': ood_fpr_95,
                                        'score_ood': score_ood.mean().item(),
                                        }
            if name not in scores:
                scores[name] = {}
            scores[name][type_name] = score_ood
            ood_auc_mean.append(ood_auc)
    mean_auc = int((sum(ood_auc_mean) / len(ood_auc_mean))*1000)
    checkpoint_path = pathlib.Path(checkpoints)
    checkpoint_path.mkdir(parents=True, exist_ok=True)
    model_path = checkpoint_path / f"{method}_{dataset}_{encoder}.pt"
    torch.save({
        "config": conf,
        "model": ood_model.state_dict(),
        "results": results,
        "method": method,
        "encoder": encoder,
        "dataset": dataset,
    }, model_path)

    with open(checkpoint_path/f"{method}_{dataset}_{encoder}_scores.pkl", 'wb') as f:
        pickle.dump(scores, f)
    return results

def objective(method, encoder, dataset, method_configs, checkpoints, device):
    # features = pathlib.Path(r"H:\arty\data\features_opt")
    features = pathlib.Path("/mnt/data/arty/data/features_opt")
    features_data = {}
    all_pkl = list(features.rglob("*.pkl"))
    for path in all_pkl:
        parts = path.parts[len(features.parts):]
        tmp = features_data
        for p in parts[:-2]:
            if p not in tmp:
                tmp[p] = {}
            tmp = tmp[p]
        else:
            tmp[parts[-2]] = path
    if pathlib.Path(f"{checkpoints}/{method}_{encoder}_{dataset}_scores.pkl").exists():
        return
    results = {}
    conf = method_configs[method]
    res = run(conf, features_data, encoder, dataset, method, device, checkpoints=checkpoints)
    if method not in results:
        results[method] = {}
    if encoder not in results[method]:
        results[method][encoder] = {}
    if dataset not in results[method][encoder]:
        results[method][encoder][dataset] = {}
    results[method][encoder][dataset]["id"] = {
        "dataset": dataset+"_test",
        "metrics": {
            "AUC": float(res['id']['AUC']),
            "FPR_95": float(res['id']['FPR_95']),
            "score_id": float(res['id']['score_id']),
            "score_ref": float(res['id']['score_ref']),
            "loss": float(res['id']['loss']),
        },
    }
    means = {}  
    for type_name, datasets in res.items():
        if type_name == 'id':
            continue
        if type_name not in results[method][encoder]:
            results[method][encoder][dataset][type_name] = []
        for dataset_name, res in datasets.items():
            results[method][encoder][dataset][type_name].append({
                "dataset": dataset_name,
                "metrics": {
                    "AUC": float(res['AUC']),
                    "FPR_95": float(res['FPR_95']),
                    "score": float(res['score_ood']),
                },
            })
            if type_name not in means:
                means[type_name] = {}
            if "AUC" not in means[type_name]:
                means[type_name]["AUC"] = []
            if "FPR_95" not in means[type_name]:
                means[type_name]["FPR_95"] = []
            if "score" not in means[type_name]:
                means[type_name]["score"] = []
            
            means[type_name]["AUC"].append(float(res['AUC']))
            means[type_name]["FPR_95"].append(float(res['FPR_95']))
            means[type_name]["score"].append(float(res['score_ood']))

    pathlib.Path("results").mkdir(parents=True, exist_ok=True)
    with open(f"results/{method}_{encoder}_{dataset}.yaml", "w") as f:
        yaml.dump(results, f, sort_keys=False)

def main(checkpoint = "results_v3"):

    # datasets = ['imagenet', 'imagenet200', 'cifar10', 'cifar100', 'covid', 'mnist']
    datasets = ['imagenet_sub', 'imagenet', 'imagenet200', 'cifar10', 'cifar100']
    general_encoders = ['dino', 'dinov2', 'vit', 'clip']
    encoders = ['repvgg', 'resnet50d', 'swin', 'deit', 'swin_t', 'vit_b16']
    # encoders = ['dinov2', 'vit', 'clip']
    open_ood_encoders = ['resnet18_32x32_cifar10_open_ood', 'resnet18_32x32_cifar100_open_ood', 'resnet18_224x224_imagenet200_open_ood', 'resnet50_224x224_imagenet_open_ood']
    open_ood_datasets = ['cifar10', 'cifar100', 'imagenet200', 'imagenet']
    methods = ['subVPSDE', 'VESDE', 'VPSDE']
    # methods = ['Residual']
    jobs = []
    # train_config = {
    #     'n_epochs': 300,
    #     'bottleneck_channels': 768,
    #     'num_res_blocks': 10,
    #     'time_embed_dim': 512,
    #     'dropout': 0.3,
    #     'lr':5e-5,
    #     'beta1': 0.9,
    #     'beta2': 0.999,
    #     'eps': 1e-8,
    #     'weight_decay': 0,
    #     'continuous': True,
    #     'reduce_mean': True,
    #     'likelihood_weighting': False,
    # }
    train_config = {
        'n_epochs': 500,
        'bottleneck_channels': 512,
        'num_res_blocks': 5,
        'time_embed_dim': 512,
        'dropout': 0.0,
        'lr':5e-5,
        'beta1': 0.9,
        'beta2': 0.999,
        'eps': 1e-8,
        'weight_decay': 0,
        'continuous': True,
        'reduce_mean': True,
        'likelihood_weighting': False,
    }


    method_configs = {
        'Residual':
            {
                'dims': None,
            },
        'VESDE':
            {
                **train_config,
                'sigma_min': 0.05,
                'sigma_max': 30,
            },
        'VPSDE':
            {
                **train_config,
                'beta_min': 0.5,
                'beta_max': 15,
            },
        'subVPSDE':
            {
                **train_config,
                'beta_min': 0.5,
                'beta_max': 15,
            },
    }

    for m in methods:
        for e in encoders+general_encoders:
            for d in datasets:
                jobs.append((m, e, d, method_configs, checkpoint))

        for e, d in zip(open_ood_encoders, open_ood_datasets):
            jobs.append((m, e, d, method_configs, checkpoint))

    
    gpu_nodes = [0, 1, 2, 3] * 2
    random.shuffle(jobs)
    print(f'Running {len(jobs)} jobs...')
    ops_utils.parallelize(objective, jobs, gpu_nodes, verbose=True, timeout=60*60*24)


if __name__ == '__main__':
    mp.freeze_support()
    main()



import math
import torch
import torch.nn as nn
from einops import repeat


def timestep_embedding(timesteps, dim, max_period=10000, repeat_only=False):
    """
    Create sinusoidal timestep embeddings.
    :param timesteps: a 1-D Tensor of N indices, one per batch element.
                      These may be fractional.
    :param dim: the dimension of the output.
    :param max_period: controls the minimum frequency of the embeddings.
    :param repeat_only: if True, only repeat the timesteps without embedding.
    :return: an [N x dim] Tensor of positional embeddings.
    """
    if not repeat_only:
        half = dim // 2
        freqs = torch.exp(
            -math.log(max_period) * torch.arange(start=0, end=half, dtype=torch.float32) / half
        ).to(device=timesteps.device)
        args = timesteps[:, None].float() * freqs[None]
        embedding = torch.cat([torch.cos(args), torch.sin(args)], dim=-1)
        if dim % 2:
            embedding = torch.cat([embedding, torch.zeros_like(embedding[:, :1])], dim=-1)
    else:
        embedding = repeat(timesteps, 'b -> b d', d=dim)
    return embedding


def zero_module(module):
    """
    Zero out the parameters of a module and return it.
    :param module: the module to zero out.
    :return: the zeroed module.
    """
    for p in module.parameters():
        p.detach().zero_()
    return module


class ResBlock(nn.Module):
    """
    A residual block that can optionally change the number of channels.
    :param channels: the number of input channels.
    :param mid_channels: the number of middle channels.
    :param emb_channels: the number of timestep embedding channels.
    :param dropout: the rate of dropout.
    :param use_context: if True, use context conditioning.
    :param context_channels: the number of context channels.
    """

    def __init__(
        self,
        channels,
        mid_channels,
        emb_channels,
        dropout,
        use_context=False,
        context_channels=512
    ):
        super().__init__()
        self.channels = channels
        self.emb_channels = emb_channels
        self.dropout = dropout

        self.in_layers = nn.Sequential(
            nn.LayerNorm(channels),
            nn.SiLU(),
            nn.Linear(channels, mid_channels, bias=True),
        )

        self.emb_layers = nn.Sequential(
            nn.SiLU(),
            nn.Linear(emb_channels, mid_channels, bias=True),
        )

        self.out_layers = nn.Sequential(
            nn.LayerNorm(mid_channels),
            nn.SiLU(),
            nn.Dropout(p=dropout),
            zero_module(
                nn.Linear(mid_channels, channels, bias=True)
            ),
        )

        self.use_context = use_context
        if use_context:
            self.context_layers = nn.Sequential(
                nn.SiLU(),
                nn.Linear(context_channels, mid_channels, bias=True),
            )

    def forward(self, x, emb, context):
        """
        Forward pass of the residual block.
        :param x: the input tensor.
        :param emb: the timestep embedding tensor.
        :param context: the context tensor.
        :return: the output tensor.
        """
        h = self.in_layers(x)
        emb_out = self.emb_layers(emb)
        if self.use_context:
            context_out = self.context_layers(context)
            h = h + emb_out + context_out
        else:
            h = h + emb_out
        h = self.out_layers(h)
        return x + h


class SimpleMLP(nn.Module):
    """
    The full skip network with timestep embedding.
    :param channels: channels in the input Tensor.
    :param time_embed_dim: dimension of the timestep embedding.
    :param bottleneck_channels: base channel count for the model.
    :param num_res_blocks: number of residual blocks per downsample.
    :param dropout: the rate of dropout.
    :param use_context: if True, use context conditioning.
    :param context_channels: the number of context channels.
    """

    def __init__(
        self,
        channels,
        time_embed_dim=512,
        bottleneck_channels=512,
        num_res_blocks=5,
        dropout=0,
        use_context=False,
        context_channels=512
    ):
        super().__init__()

        self.image_size = 1
        self.channels = channels
        self.num_res_blocks = num_res_blocks
        self.dropout = dropout

        self.time_embed = nn.Sequential(
            nn.Linear(channels, time_embed_dim),
            nn.SiLU(),
            nn.Linear(time_embed_dim, time_embed_dim),
        )

        self.input_proj = nn.Linear(channels, channels)

        res_blocks = []
        for i in range(num_res_blocks):
            res_blocks.append(ResBlock(
                channels,
                bottleneck_channels,
                time_embed_dim,
                dropout,
                use_context=use_context,
                context_channels=context_channels
            ))

        self.res_blocks = nn.ModuleList(res_blocks)

        self.out = nn.Sequential(
            nn.LayerNorm(channels, eps=1e-6),
            nn.SiLU(),
            zero_module(nn.Linear(channels, channels, bias=True)),
        )

    def forward(self, x, timesteps=None, context=None, y=None, **kwargs):
        """
        Apply the model to an input batch.
        :param x: an [N x C x ...] Tensor of inputs.
        :param timesteps: a 1-D batch of timesteps.
        :param context: conditioning plugged in via crossattn.
        :param y: an [N] Tensor of labels, if class-conditional.
        :return: an [N x C x ...] Tensor of outputs.
        """
        x = x.squeeze()
        x = self.input_proj(x)
        t_emb = timestep_embedding(timesteps, self.channels, repeat_only=False)
        emb = self.time_embed(t_emb)

        for block in self.res_blocks:
            x = block(x, emb, context)
        x = self.out(x)
        return x



import timm
import torch
import torchvision
import pathlib, requests, pathlib, gdown, zipfile
from torchvision.datasets import ImageFolder
from ood_detectors.list_dataset import ImageFilelist
import torch.nn as nn
import torch.nn.functional as F
from torchvision.models.resnet import BasicBlock, ResNet, Bottleneck

def download(url, filename):
    data = requests.get(url)
    with open(filename, 'wb') as f:
        f.write(data.content)

def load_model(obj, builder, model_dir_path, weights):
    model_dir_path = pathlib.Path(model_dir_path)
    model_dir_path.mkdir(exist_ok=True, parents=True)
    if weights not in obj.WEIGTHS:
        raise ValueError(f"weights should be one of {obj.WEIGTHS.keys()}")
    model = builder(weights=obj.WEIGTHS[weights])
    return model

TRANSFORM_IMAGENET1K_V1 = torchvision.transforms.Compose([
    torchvision.transforms.Resize((224, 224)),
    torchvision.transforms.ToTensor(),
    torchvision.transforms.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225))
])

TRANSFORM_IMAGENET1K_SWAG_E2E_V1_384 = torchvision.transforms.Compose([
    torchvision.transforms.Resize((384, 384)),
    torchvision.transforms.ToTensor(),
    torchvision.transforms.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225))
])
TRANSFORM_IMAGENET1K_SWAG_E2E_V1_512 = torchvision.transforms.Compose([
    torchvision.transforms.Resize((512, 512)),
    torchvision.transforms.ToTensor(),
    torchvision.transforms.Normalize(mean=(0.485, 0.456, 0.406), std=(0.229, 0.224, 0.225))
])

TRANSFORMS = {
    "ViT_B_16_IMAGENET1K_V1": TRANSFORM_IMAGENET1K_V1,
    "ViT_B_16_IMAGENET1K_SWAG_E2E_V1": TRANSFORM_IMAGENET1K_SWAG_E2E_V1_384,
    "ViT_B_16_IMAGENET1K_SWAG_LINEAR_V1": TRANSFORM_IMAGENET1K_V1,
    "ViT_B_32_IMAGENET1K_V1": TRANSFORM_IMAGENET1K_V1,
    "ViT_L_16_IMAGENET1K_V1": TRANSFORM_IMAGENET1K_V1,
    "ViT_L_16_IMAGENET1K_SWAG_E2E_V1": TRANSFORM_IMAGENET1K_SWAG_E2E_V1_512,
    "ViT_L_16_IMAGENET1K_SWAG_LINEAR_V1": TRANSFORM_IMAGENET1K_V1,
    "ViT_L_32_IMAGENET1K_V1": TRANSFORM_IMAGENET1K_V1,
}

class TimmModel(torch.nn.Module):
    def __init__(self, model_name: str):
        super().__init__()
        self.name = model_name
        self.model = timm.create_model(model_name, pretrained=True, num_classes=0)
        data_config = timm.data.resolve_model_data_config(self.model)   
        self.transform = timm.data.create_transform(**data_config, is_training=True, no_aug=True)


    def forward(self, x):
        x = self.model(x)
        return x
    
    def features(self, x):
        x = self.model(x)
        return x



class BASE_ViT(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.transform = TRANSFORMS[self.name]

    def forward(self, x):
        x = self.model(x)
        return x
    
    def features(self, x):
        x = self.model._process_input(x)
        n = x.shape[0]

        # Expand the class token to the full batch
        batch_class_token = self.model.class_token.expand(n, -1, -1)
        x = torch.cat([batch_class_token, x], dim=1)

        x = self.model.encoder(x)

        # Classifier "token" as used by standard language architectures
        x = x[:, 0]
        return x

class ViT_B_16(BASE_ViT):
    WEIGTHS = {
        "IMAGENET1K_V1": torchvision.models.ViT_B_16_Weights.IMAGENET1K_V1,
        "IMAGENET1K_SWAG_E2E_V1": torchvision.models.ViT_B_16_Weights.IMAGENET1K_SWAG_E2E_V1,
        "IMAGENET1K_SWAG_LINEAR_V1": torchvision.models.ViT_B_16_Weights.IMAGENET1K_SWAG_LINEAR_V1,
    }
    def __init__(self, model_dir_path='checkpoints', weights='IMAGENET1K_V1'):
        self.name = f'ViT_B_16_{weights}'
        super().__init__()
        self.model = load_model(self, torchvision.models.vit_b_16, model_dir_path, weights)
        self.name = f'vit_b_16'


    def features(self, x):
        # Reshape and permute the input tensor
        x = self.model._process_input(x)
        n = x.shape[0]

        # Expand the class token to the full batch
        batch_class_token = self.model.class_token.expand(n, -1, -1)
        x = torch.cat([batch_class_token, x], dim=1)

        x = self.model.encoder(x)

        # Classifier "token" as used by standard language architectures
        x = x[:, 0]
        return x

 
class ViT_B_32(BASE_ViT):
    WEIGTHS = {
        "IMAGENET1K_V1": torchvision.models.ViT_B_32_Weights.IMAGENET1K_V1,
    }
    def __init__(self, model_dir_path='checkpoints', weights='IMAGENET1K_V1'):
        self.name = f'ViT_B_32_{weights}'
        super().__init__()
        self.model = load_model(self, torchvision.models.vit_b_32, model_dir_path, weights)
        self.name = f'vit_b_32'

  

class ViT_L_16(BASE_ViT):
    WEIGTHS = {
        "IMAGENET1K_V1": torchvision.models.ViT_L_16_Weights.IMAGENET1K_V1,
        "IMAGENET1K_SWAG_E2E_V1": torchvision.models.ViT_L_16_Weights.IMAGENET1K_SWAG_E2E_V1,
        "IMAGENET1K_SWAG_LINEAR_V1": torchvision.models.ViT_L_16_Weights.IMAGENET1K_SWAG_LINEAR_V1,
    }
    def __init__(self, model_dir_path='checkpoints', weights='IMAGENET1K_V1'):
        self.name = f'ViT_L_16_{weights}'
        super().__init__()
        self.model = load_model(self, torchvision.models.vit_l_16, model_dir_path, weights)
        self.name = f'vit_l_16'

   

class VIT_L_32(BASE_ViT):
    WEIGTHS = {
        "IMAGENET1K_V1": torchvision.models.ViT_L_32_Weights.IMAGENET1K_V1,
    }
    def __init__(self, model_dir_path='checkpoints', weights='IMAGENET1K_V1'):
        super().__init__()
        self.name = f'ViT_L_32_{weights}'
        self.model = load_model(self, torchvision.models.vit_l_32, model_dir_path, weights)
        self.name = f'vit_l_32'

class Dino_ViT_B_16(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.name = 'dino_vit_b_16'
        self.model = torch.hub.load('facebookresearch/dino:main', 'dino_vitb16')
        
        self.transform = torchvision.transforms.Compose([
            torchvision.transforms.Resize((224, 224)),
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
        ])

    def forward(self, x):
        x = self.model(x)
        return x
    
    def features(self, x):
        x = self.model(x)
        return x

class DinoV2_ViT_B_14(torch.nn.Module):
    def __init__(self):
        super().__init__()
        self.name = 'dinov2_vit_b_14'
        self.model = torch.hub.load('facebookresearch/dinov2', 'dinov2_vitb14')
        
        self.transform = torchvision.transforms.Compose([
            torchvision.transforms.Resize((224, 224)),
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
        ])

    def forward(self, x):
        x = self.model(x)
        return x
    
    def features(self, x):
        x = self.model(x)
        return x
 
class ViT_P16_21k(torch.nn.Module):
    def __init__(self, model_dir_path='checkpoints'):
        from mmpretrain.apis import init_model
        import mmengine
        super().__init__()
        self.name = 'vit_p_16_21k'
        model_dir = pathlib.Path(model_dir_path)
        weights = model_dir/'vit-base-p16_in21k-pre-3rdparty_ft-64xb64_in1k-384_20210928-98e8652b.pth'
        if not model_dir.exists():
            model_dir.mkdir(exist_ok=True)
        if not weights.exists():
            url = 'https://download.openmmlab.com/mmclassification/v0/vit/finetune/vit-base-p16_in21k-pre-3rdparty_ft-64xb64_in1k-384_20210928-98e8652b.pth'
            download(url, weights)
        self.transform = torchvision.transforms.Compose([
            torchvision.transforms.Resize((384, 384)),
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize((0.5, 0.5, 0.5), (0.5, 0.5, 0.5)),
        ])
        cfg = mmengine.Config(cfg_dict=dict(model=dict(
            type='ImageClassifier',
            backbone=dict(
                type='VisionTransformer',
                arch='b',
                img_size=384,
                patch_size=16,
                drop_rate=0.1,
                init_cfg=[
                    dict(
                        type='Kaiming',
                        layer='Conv2d',
                        mode='fan_in',
                        nonlinearity='linear')
                ]),
            neck=None,
            head=dict(
                type='VisionTransformerClsHead',
                num_classes=1000,
                in_channels=768,
                loss=dict(
                    type='LabelSmoothLoss', label_smooth_val=0.1,
                    mode='classy_vision'),
            )))
        )

        self.model = init_model(cfg, str(weights), 0)
        self.device = 'cpu'

    def to(self, device):
        self.model.to(device)
        self.device = device
        return self

    def forward(self, x):
        x = self.model(x)
        return x
    
    def features(self, x):
        x = self.model.backbone(x)[0]
        return x
    
class ClipVisionModel(torch.nn.Module):
    def __init__(self):
        import clip
        super().__init__()
        self.name = 'CLIP'
        self.model, self.transform = clip.load("ViT-B/32", device='cpu')
        self.device = 'cpu'

    def to(self, device):
        self.model.visual.to(device) # Move the visual model to the specified device we don't need the text model
        self.device = device
        return self

    def forward(self, x):
        x = self.model.encode_image(x)
        return x

    def features(self, x):
        x = self.model.encode_image(x)
        return x

class _BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        super().__init__()
        self.conv1 = nn.Conv2d(in_planes,
                               planes,
                               kernel_size=3,
                               stride=stride,
                               padding=1,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes,
                               planes,
                               kernel_size=3,
                               stride=1,
                               padding=1,
                               bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion * planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes,
                          self.expansion * planes,
                          kernel_size=1,
                          stride=stride,
                          bias=False), nn.BatchNorm2d(self.expansion * planes))

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class ResNet18_32x32(nn.Module):
    def __init__(self, block=_BasicBlock, num_blocks=None, num_classes=10):
        super().__init__()
        if num_blocks is None:
            num_blocks = [2, 2, 2, 2]
        self.in_planes = 64

        self.conv1 = nn.Conv2d(3,
                               64,
                               kernel_size=3,
                               stride=1,
                               padding=1,
                               bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.layer1 = self._make_layer(block, 64, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(block, 128, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(block, 256, num_blocks[2], stride=2)
        self.layer4 = self._make_layer(block, 512, num_blocks[3], stride=2)
        # self.avgpool = nn.AvgPool2d(4)
        self.avgpool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Linear(512 * block.expansion, num_classes)
        self.feature_size = 512 * block.expansion

    def _make_layer(self, block, planes, num_blocks, stride):
        strides = [stride] + [1] * (num_blocks - 1)
        layers = []
        for stride in strides:
            layers.append(block(self.in_planes, planes, stride))
            self.in_planes = planes * block.expansion
        return nn.Sequential(*layers)
    
    def features(self, x):
        feature1 = F.relu(self.bn1(self.conv1(x)))
        feature2 = self.layer1(feature1)
        feature3 = self.layer2(feature2)
        feature4 = self.layer3(feature3)
        feature5 = self.layer4(feature4)
        feature5 = self.avgpool(feature5)
        feature = feature5.view(feature5.size(0), -1)
        return feature

    def forward(self, x):
        return self.fc(self.features(x))
    
class ResNet18_224x224(ResNet):
    def __init__(self,
                 block=BasicBlock,
                 layers=[2, 2, 2, 2],
                 num_classes=1000):
        super().__init__(block=block,
                                               layers=layers,
                                               num_classes=num_classes)
        self.feature_size = 512

    def features(self, x):
        feature1 = self.relu(self.bn1(self.conv1(x)))
        feature1 = self.maxpool(feature1)
        feature2 = self.layer1(feature1)
        feature3 = self.layer2(feature2)
        feature4 = self.layer3(feature3)
        feature5 = self.layer4(feature4)
        feature5 = self.avgpool(feature5)
        feature = feature5.view(feature5.size(0), -1)
        return feature
        
    def forward(self, x):
        return self.fc(self.features(x))
    
class ResNet50(ResNet):
    def __init__(self,
                 block=Bottleneck,
                 layers=[3, 4, 6, 3],
                 num_classes=1000):
        super(ResNet50, self).__init__(block=block,
                                       layers=layers,
                                       num_classes=num_classes)
        self.feature_size = 2048

    def features(self, x):
        feature1 = self.relu(self.bn1(self.conv1(x)))
        feature1 = self.maxpool(feature1)
        feature2 = self.layer1(feature1)
        feature3 = self.layer2(feature2)
        feature4 = self.layer3(feature3)
        feature5 = self.layer4(feature4)
        feature5 = self.avgpool(feature5)
        feature = feature5.view(feature5.size(0), -1)
        return feature
    
    def forward(self, x):
        return self.fc(self.features(x))

class Swin_T(nn.Module):
    def __init__(self):
        super().__init__()
        self.name = "swin_t"
        self.model = torchvision.models.swin_t(weights=torchvision.models.Swin_T_Weights.IMAGENET1K_V1)
        self.transform = torchvision.transforms.Compose([
            torchvision.transforms.Resize((224, 224)),
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize((0.485, 0.456, 0.406), (0.229, 0.224, 0.225)),
        ])

    def features(self, x):
        x = self.model.features(x)
        x = self.model.norm(x)
        x = self.model.permute(x)
        x = self.model.avgpool(x)
        x = self.model.flatten(x)
        return x

    def forward(self, x):
        x = self.model.features(x)
        return self.model.head(x)


normalization_dict = {
    'cifar10': [[0.4914, 0.4822, 0.4465], [0.2470, 0.2435, 0.2616]],
    'cifar100': [[0.5071, 0.4867, 0.4408], [0.2675, 0.2565, 0.2761]],
    'imagenet': [[0.485, 0.456, 0.406], [0.229, 0.224, 0.225]],
}

download_id_open_ood = {
    'cifar10_res18_v1.5': '1byGeYxM_PlLjT72wZsMQvP6popJeWBgt',
    'cifar100_res18_v1.5': '1s-1oNrRtmA0pGefxXJOUVRYpaoAML0C-',
    'imagenet200_res18_v1.5': '1ddVmwc8zmzSjdLUO84EuV4Gz1c7vhIAs',
    'imagenet_res50_v1.5': '15PdDMNRfnJ7f2oxW6lI-Ge4QJJH3Z0Fy',
    'datalist': '1XKzBdWCqg3vPoj-D32YixJyJJ0hL63gP',
    'usps': '1KhbWhlFlpFjEIb4wpvW0s9jmXXsHonVl',
    'cifar100': '1PGKheHUsf29leJPPGuXqzLBMwl8qMF8_',
    'cifar10': '1Co32RiiWe16lTaiOU6JMMnyUYS41IlO1',
    'cifar10c': '170DU_ficWWmbh6O2wqELxK9jxRiGhlJH',
    'cinic10': '190gdcfbvSGbrRK6ZVlJgg5BqqED6H_nn',
    'svhn': '1DQfc11HOtB1nEwqS4pWUFp8vtQ3DczvI',
    'fashionmnist': '1nVObxjUBmVpZ6M0PPlcspsMMYHidUMfa',
    'cifar100c': '1MnETiQh9RTxJin2EHeSoIAJA28FRonHx',
    'mnist': '1CCHAGWqA1KJTFFswuF9cbhmB-j98Y1Sb',
    'tin': '1PZ-ixyx52U989IKsMA2OT-24fToTrelC',
    'texture': '1OSz1m3hHfVWbRdmMwKbUzoU8Hg9UKcam',
    'notmnist': '16ueghlyzunbksnc_ccPgEAloRW9pKO-K',
    'places365': '1Ec-LRSTf6u5vEctKX9vRp9OA6tqnJ0Ay',
    'places': '1fZ8TbPC4JGqUCm-VtvrmkYxqRNp2PoB3',
    'sun': '1ISK0STxWzWmg-_uUr4RQ8GSLFW7TZiKp',
    'species_sub': '1-JCxDx__iFMExkYRMylnGJYTPvyuX6aq',
    'imagenet_1k': '1i1ipLDFARR-JZ9argXd2-0a6DXwVhXEj',
    'ssb_hard': '1PzkA-WGG8Z18h0ooL_pDdz9cO-DCIouE',
    'ninco': '1Z82cmvIB0eghTehxOGP5VTdLt7OD3nk6',
    'imagenet_v2': '1akg2IiE22HcbvTBpwXQoD7tgfPCdkoho',
    'imagenet_r': '1EzjMN2gq-bVV7lg-MEAdeuBuz-7jbGYU',
    'imagenet_c': '1JeXL9YH4BO8gCJ631c5BHbaSsl-lekHt',
    'imagenet_o': '1S9cFV7fGvJCcka220-pIO9JPZL1p1V8w',
    'openimage_o': '1VUFXnB_z70uHfdgJG2E_pjYOcEgqM7tE',
    'inaturalist': '1zfLfMvoUD0CUlKNnkk7LgxZZBnTBipdj',
    'actmed': '1tibxL_wt6b3BjliPaQ2qjH54Wo4ZXWYb',
    'ct': '1k5OYN4inaGgivJBJ5L8pHlopQSVnhQ36',
    'hannover': '1NmqBDlcA1dZQKOvgcILG0U1Tm6RP0s2N',
    'xraybone': '1ZzO3y1-V_IeksJXEvEfBYKRoQLLvPYe9',
    'bimcv': '1nAA45V6e0s5FAq2BJsj9QH5omoihb7MZ',
}

data_list_open_ood = [
    'usps',
    'cifar100',
    'cifar10',
    'cifar10c',
    'cinic10',
    'svhn',
    'fashionmnist',
    'cifar100c',
    'mnist',
    'tin',
    'texture',
    'notmnist',
    'places365',
    'places',
    'sun',
    'species_sub',
    'imagenet_1k',
    'ssb_hard',
    'ninco',
    'imagenet_v2',
    'imagenet_r',
    'imagenet_c',
    'imagenet_o',
    'openimage_o',
    'inaturalist',
    'actmed',
    'ct',
    'hannover',
    'xraybone',
    'bimcv',
]

AVAILABLE_ENCODERS = ['repvgg', 'resnet50d', 'swin', 'deit', 'dino', 'dinov2', 'vit', 'vit_b16', 'clip', 'swin_t', 'resnet18_32x32_cifar10_open_ood', 'resnet18_32x32_cifar100_open_ood', 'resnet18_224x224_imagenet200_open_ood', 'resnet50_224x224_imagenet_open_ood']

def download_and_extract(hash, checkpoint_path):
    gdown.download(id=hash, output=str(checkpoint_path / f"{hash}.zip") )
    with zipfile.ZipFile(checkpoint_path/f"{hash}.zip", 'r') as zip_ref:
        zip_ref.extractall(checkpoint_path)
    (checkpoint_path/f"{hash}.zip").unlink()


def get_encoder(name):
    name = name.lower()
    if name in ['repvgg', 'repvgg_b3']:
        return TimmModel('repvgg_b3')
    elif name in ['resnet50d', 'res50d', 'resnet50']:
        return TimmModel('resnet50d')
    elif name in ['swin', 'swin_base_patch4_window7_224']:
        return TimmModel('swin_base_patch4_window7_224')
    elif name in ['deit', 'deit_base_patch16_224']:
        return TimmModel('deit_base_patch16_224')
    elif name in ['dino', 'dino_vit_b_16']:
        return Dino_ViT_B_16()
    elif name in ['dinov2', 'dinov2_vit_b_14']:
        return DinoV2_ViT_B_14()
    elif name in ['vit','vit_p16']:
        return ViT_P16_21k()
    elif name in ['vit_b16']:
        return ViT_B_16()
    elif name in ['swin_t']:
        return Swin_T()
    elif name in ['clip']:
        return ClipVisionModel()
    elif name in ['resnet18_32x32_cifar10_open_ood']:
        model = ResNet18_32x32()
        checkpoint = pathlib.Path('checkpoints/')
        checkpoint.mkdir(exist_ok=True)
        model_path = checkpoint/'cifar10_resnet18_32x32_base_e100_lr0.1_default/s0/best_epoch96_acc0.9470.ckpt'
        if not model_path.exists():
            download_and_extract(download_id_open_ood['cifar10_res18_v1.5'], checkpoint)
        model.load_state_dict(torch.load(model_path))
        model.transform = torchvision.transforms.Compose([
            torchvision.transforms.Resize((32, 32)),
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize(*normalization_dict['cifar10'])
        ])
        model.name = 'resnet18_32x32_cifar10_open_ood'
        return model
    elif name in ['resnet18_32x32_cifar100_open_ood']:
        model = ResNet18_32x32(num_classes=100)
        checkpoint = pathlib.Path('checkpoints/')
        checkpoint.mkdir(exist_ok=True)
        model_path = checkpoint/'cifar100_resnet18_32x32_base_e100_lr0.1_default/s0/best_epoch99_acc0.7810.ckpt'
        if not model_path.exists():
            download_and_extract(download_id_open_ood['cifar100_res18_v1.5'], checkpoint)
        model.load_state_dict(torch.load(model_path))
        model.transform = torchvision.transforms.Compose([
            torchvision.transforms.Resize((32, 32)),
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize(*normalization_dict['cifar100'])
        ])
        model.name = 'resnet18_32x32_cifar100_open_ood'
        return model
    elif name in ['resnet18_224x224_imagenet200_open_ood']:
        model = ResNet18_224x224(num_classes=200)
        checkpoint = pathlib.Path('checkpoints/')
        checkpoint.mkdir(exist_ok=True)
        model_path = checkpoint/'imagenet200_resnet18_224x224_base_e90_lr0.1_default/s0/best_epoch89_acc0.8500.ckpt'
        if not model_path.exists():
            download_and_extract(download_id_open_ood['imagenet200_res18_v1.5'], checkpoint)
        model.load_state_dict(torch.load(model_path))
        model.transform = torchvision.transforms.Compose([
            torchvision.transforms.Resize((256, 256)),
            torchvision.transforms.CenterCrop(224),
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize(*normalization_dict['imagenet'])
        ])
        model.name = 'resnet18_224x224_imagenet200_open_ood'
        return model
    elif name in ['resnet50_224x224_imagenet_open_ood']:
        model = ResNet50(num_classes=1000)
        checkpoint = pathlib.Path('checkpoints/')
        checkpoint.mkdir(exist_ok=True)
        model_path = checkpoint/'imagenet_resnet50_base_e30_lr0.001_randaugment-2-9/s0/best_epoch27_acc0.7668.ckpt'
        if not model_path.exists():
            download_and_extract(download_id_open_ood['imagenet_res50_v1.5'], checkpoint)
        model.load_state_dict(torch.load(model_path))
        model.transform = torchvision.transforms.Compose([
            torchvision.transforms.Resize((256, 256)),
            torchvision.transforms.CenterCrop(224),
            torchvision.transforms.ToTensor(),
            torchvision.transforms.Normalize(*normalization_dict['imagenet'])
        ])
        model.name = 'resnet50_224x224_imagenet_open_ood'
        return model
    else:
        raise ValueError(f"Encoder {name} not available. Available encoders: {AVAILABLE_ENCODERS}")
    
def get_open_ood_datalist(checkpoint_path=pathlib.Path('checkpoints/')):
    if not (checkpoint_path/'benchmark_imglist').exists():
        download_and_extract(download_id_open_ood['datalist'], checkpoint_path)

def get_open_ood_datasets(checkpoint_path=pathlib.Path('checkpoints/')):
    for dataset in data_list_open_ood:
        if not (checkpoint_path/dataset).exists():
            download_and_extract(download_id_open_ood[dataset], checkpoint_path)

def build_dataset(data_root, transform, img_list=None):
    if img_list is not None:
        dataset = ImageFilelist(data_root, img_list, transform)
    else:
        dataset = ImageFolder(data_root, transform, allow_empty=True)
    return dataset

def imagenet_ood_datasets(data_root, transforms, image_list_root):
    datasets = {}
    datasets['id'] = {}
    datasets['nearood'] = {}
    datasets['farood'] = {}
    datasets['csid'] = {}
    datasets['id']['train'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/train_imagenet.txt')
    datasets['id']['val'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/val_imagenet.txt')
    datasets['id']['test'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/test_imagenet.txt')
    datasets['nearood']['ssb_hard'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/test_ssb_hard.txt')
    datasets['nearood']['ninco'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/test_ninco.txt')
    datasets['nearood']['imagenet-o'] = build_dataset(f'{data_root}/imagenet-o/', transforms, img_list="all")
    datasets['farood']['textures'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/test_textures.txt')
    datasets['farood']['inaturalist'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/test_inaturalist.txt')
    datasets['farood']['openimageo'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/test_openimage_o.txt')
    datasets['csid']['imagenetv2'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/test_imagenet_v2.txt')
    datasets['csid']['imagenetc'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/test_imagenet_c.txt')
    datasets['csid']['imagenetr'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/test_imagenet_r.txt')
    return datasets
    

def imagenet200_ood_datasets(data_root, transforms, image_list_root):
    datasets = {}
    datasets['id'] = {}
    datasets['nearood'] = {}
    datasets['farood'] = {}
    datasets['csid'] = {}
    datasets['id']['train'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet200/train_imagenet200.txt')
    datasets['id']['val'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet200/val_imagenet200.txt')
    datasets['id']['test'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet200/test_imagenet200.txt')
    datasets['nearood']['ssb_hard'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/test_ssb_hard.txt')
    datasets['nearood']['ninco'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/test_ninco.txt')
    datasets['nearood']['imagenet-o'] = build_dataset(f'{data_root}/imagenet-o/', transforms, img_list="all")
    datasets['farood']['textures'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/test_textures.txt')
    datasets['farood']['inaturalist'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/test_inaturalist.txt')
    datasets['farood']['openimageo'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/test_openimage_o.txt')
    datasets['csid']['imagenetv2'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/test_imagenet_v2.txt')
    datasets['csid']['imagenetc'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/test_imagenet_c.txt')
    datasets['csid']['imagenetr'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/imagenet/test_imagenet_r.txt')
    return datasets


def cifar10_ood_datasets(data_root, transforms, image_list_root):
    datasets = {}
    datasets['id'] = {}
    datasets['nearood'] = {}
    datasets['farood'] = {}
    datasets['csid'] = {}
    datasets['id']['train'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar10/train_cifar10.txt')
    datasets['id']['val'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar10/val_cifar10.txt')
    datasets['id']['test'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar10/test_cifar10.txt')
    datasets['nearood']['cifar100'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar10/test_cifar100.txt')
    datasets['nearood']['tin'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar10/test_tin.txt')
    datasets['farood']['mnist'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar10/test_mnist.txt')
    datasets['farood']['svhn'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar10/test_svhn.txt')
    datasets['farood']['texture'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar10/test_texture.txt')
    datasets['farood']['place365'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar10/test_places365.txt')
    datasets['csid']['cinic10'] = build_dataset(f"{data_root}/cinic10/val", transforms, img_list="all")
    return datasets

def cifar100_ood_datasets(data_root, transforms, image_list_root):
    datasets = {}
    datasets['id'] = {}
    datasets['nearood'] = {}
    datasets['farood'] = {}
    datasets['csid'] = {}
    datasets['id']['train'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar100/train_cifar100.txt')
    datasets['id']['val'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar100/val_cifar100.txt')
    datasets['id']['test'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar100/test_cifar100.txt')
    datasets['nearood']['cifar10'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar100/test_cifar10.txt')
    datasets['nearood']['tin'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar100/test_tin.txt')
    datasets['farood']['mnist'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar100/test_mnist.txt')
    datasets['farood']['svhn'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar100/test_svhn.txt')
    datasets['farood']['texture'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar100/test_texture.txt')
    datasets['farood']['places365'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/cifar100/test_places365.txt')
    datasets['csid']['cifar100c'] = build_dataset(f"{data_root}/cifar100c/", transforms, img_list="all")
    return datasets

def covid_ood_datasets(data_root, transforms, image_list_root):
    datasets = {}
    datasets['id'] = {}
    datasets['nearood'] = {}
    datasets['farood'] = {}
    datasets['csid'] = {}
    datasets['id']['train'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/covid/train_bimcv.txt')
    datasets['id']['val'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/covid/val_bimcv.txt')
    datasets['id']['test'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/covid/test_bimcv.txt')
    datasets['nearood']['ct'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/covid/test_ct.txt')
    datasets['nearood']['xraybone'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/covid/test_xraybone.txt')
    datasets['farood']['mnist'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/covid/test_mnist.txt')
    datasets['farood']['cifar10'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/covid/test_cifar10.txt')
    datasets['farood']['texture'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/covid/test_texture.txt')
    datasets['farood']['tin'] = build_dataset(f"{data_root}/tin/val/", transforms, img_list="all")
    datasets['csid']['actmed'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/covid/test_actmed.txt')
    datasets['csid']['hannover'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/covid/test_hannover.txt')
    return datasets

def mnist_ood_datasets(data_root, transforms, image_list_root):
    datasets = {}
    datasets['id'] = {}
    datasets['nearood'] = {}
    datasets['farood'] = {}
    datasets['csid'] = {}
    datasets['id']['train'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/mnist/train_mnist.txt')
    datasets['id']['val'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/mnist/val_mnist.txt')
    datasets['id']['test'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/mnist/test_mnist.txt')
    datasets['nearood']['notmnist'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/mnist/test_notmnist.txt')
    datasets['nearood']['fashionmnist'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/mnist/test_fashionmnist.txt')
    datasets['farood']['texture'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/mnist/test_texture.txt')
    datasets['farood']['cifar10'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/mnist/test_cifar10.txt')
    datasets['farood']['tin'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/mnist/test_tin.txt')
    datasets['farood']['places365'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/mnist/test_places365.txt')
    datasets['csid']['svhn'] = build_dataset(data_root, transforms, img_list=f'{image_list_root}/mnist/test_svhn.txt')
    return datasets

AVAILABLE_DATASETS = ['imagenet', 'imagenet200', 'cifar10', 'cifar100', 'covid', 'mnist']

def get_datasets(name, data_root, image_list_root, transform):
    if name == 'imagenet':
        return imagenet_ood_datasets(data_root, transform, image_list_root)
    elif name == 'imagenet200':
        return imagenet200_ood_datasets(data_root, transform, image_list_root)
    elif name == 'cifar10':
        return cifar10_ood_datasets(data_root, transform, image_list_root)
    elif name == 'cifar100':
        return cifar100_ood_datasets(data_root, transform, image_list_root)
    elif name == 'covid':
        return covid_ood_datasets(data_root, transform, image_list_root)
    elif name == 'mnist':
        return mnist_ood_datasets(data_root, transform, image_list_root)
    else:
        raise ValueError(f"Dataset {name} not available. Available datasets: {AVAILABLE_DATASETS}")
    
def extract_features(encoder, dataset, batch_size=64, num_workers=8, device='cuda'):
    dataloader = torch.utils.data.DataLoader(dataset, batch_size=batch_size, shuffle=False, num_workers=num_workers)
    features = []
    encoder.to(device)
    with torch.no_grad():
        for x, _ in dataloader:
            x = x.to(device)
            feature = encoder.features(x).cpu()
            features.append(feature)
    features = torch.cat(features)
    return features


import torch
from ood_detectors import train, losses, ood_utils
from ood_detectors.models import SimpleMLP
from ood_detectors.sde import VESDE, subVPSDE, VPSDE
import torch.optim as optim
import functools


class Likelihood:
    """Base class for likelihood function implementations for different SDEs with dependency injection.

    Args:
        sde (object): The SDE object representing the stochastic differential equation.
        optimizer (function): The optimizer function used for training the model. Default is Adam optimizer.
        model (object): The model object used for likelihood estimation. If None, a SimpleMLP model will be created.
        feat_dim (int): The feature dimension of the model. Required if model is None.

    Attributes:
        model (object): The model object used for likelihood estimation.
        sde (object): The SDE object representing the stochastic differential equation.
        optimizer (object): The optimizer object used for training the model.
        device (str): The device (CPU or GPU) on which the model is loaded.
        name (str): The name of the likelihood function.

    Methods:
        to(device): Moves the model to the specified device.
        load_state_dict(state_dict): Loads the model state from a state dictionary.
        state_dict(): Returns the model state dictionary.
        fit(dataset, n_epochs, batch_size, num_workers, update_fn, verbose): Trains the model on the given dataset.
        predict(dataset, batch_size, num_workers, verbose): Performs inference on the given dataset.
        __call__(*args, **kwargs): Calls the predict method.

    """

    def __init__(self, sde, 
                 optimizer=functools.partial(
                    optim.Adam,
                    lr=5e-5,
                    betas=(0.9, 0.999),
                    eps=1e-8,
                    weight_decay=0,
                 ), 
                 model=None, 
                 feat_dim=None,
                 ):
        self.model = model
        if model is None:
            assert feat_dim is not None, "feat_dim must be provided if model is None"
            self.model = SimpleMLP(feat_dim)
        self.sde = sde
        self.optimizer = optimizer(self.model.parameters())
        self.device = "cpu"
        self.name = f"{self.sde.__class__.__name__}_{self.model.__class__.__name__}"

    def to(self, device):
        """Moves the model to the specified device.

        Args:
            device (str): The device to which the model should be moved (cpu or cuda).

        Returns:
            self: The Likelihood object.

        """
        self.model.to(device)
        self.device = device
        return self

    def load_state_dict(self, state_dict):
        """Loads the model state from a state dictionary.

        Args:
            state_dict (dict): The state dictionary containing the model state.

        """
        self.model.load_state_dict(state_dict)

    def state_dict(self):
        """Returns the model state dictionary.

        Returns:
            dict: The model state dictionary.

        """
        return self.model.state_dict()

    def fit(self, dataset, n_epochs, batch_size, 
            num_workers=0,
            update_fn=None, 
            verbose=True,
            collate_fn=None,
            ):
        """Trains the model on the given dataset.

        Args:
            dataset (object): The dataset object containing the training data.
            n_epochs (int): The number of epochs to train the model.
            batch_size (int): The batch size for training.
            num_workers (int): The number of worker processes for data loading. Default is 0.
            update_fn (object): The update function for updating the model parameters. If None, a default update function is used.
            verbose (bool): Whether to print training progress. Default is True.

        Returns:
            dict: A dictionary containing the training loss and other metrics.

        """
        if update_fn is None:
            update_fn = losses.SDE_BF16(
                sde=self.sde,
                model=self.model,
                optimizer=self.optimizer,
            )
        else:
            update_fn = update_fn(sde=self.sde, model=self.model, optimizer=self.optimizer)
        
        if collate_fn is None and getattr(dataset, 'collate_fn', None) is not None:
            collate_fn = dataset.collate_fn

        return train.train(dataset, self.model, update_fn, n_epochs, batch_size, self.device, num_workers, verbose=verbose, collate_fn=collate_fn)

    def predict(self, dataset, batch_size, num_workers=0, verbose=True, collate_fn=None):
        """Performs inference on the given dataset.

        Args:
            dataset (object): The dataset object containing the test data.
            batch_size (int): The batch size for inference.
            num_workers (int): The number of worker processes for data loading. Default is 0.
            verbose (bool): Whether to print inference progress. Default is True.

        Returns:
            dict: A dictionary containing the likelihood scores for the test data.

        """
        likelihood_fn = ood_utils.get_likelihood_fn(self.sde)
        if collate_fn is None and getattr(dataset, 'collate_fn', None) is not None:
            collate_fn = dataset.collate_fn
        return train.inference(dataset, self.model, likelihood_fn, batch_size, self.device, num_workers, verbose=verbose, collate_fn=collate_fn)
    
    def __call__(self, *args, **kwargs):
        """Calls the predict method.

        Args:
            *args: Variable length argument list.
            **kwargs: Arbitrary keyword arguments.

        Returns:
            dict: A dictionary containing the likelihood scores.

        """
        return self.predict(*args, **kwargs)
    
def RDM_VESDE(feat_dim):
    """Creates a Likelihood object for Variational Ensemble SDE with Dependency Detection Model.

    Args:
        feat_dim (int): The feature dimension of the model.

    Returns:
        Likelihood: The Likelihood object.

    """
    return Likelihood(VESDE(sigma_min=0.05, sigma_max=30, N=1000), feat_dim=feat_dim)

def RDM_SubSDE(feat_dim):
    """Creates a Likelihood object for Sub-Variational Poisson SDE with Dependency Detection Model.

    Args:
        feat_dim (int): The feature dimension of the model.

    Returns:
        Likelihood: The Likelihood object.

    """
    return Likelihood(subVPSDE(beta_min=0.5, beta_max=15, N=1000), feat_dim=feat_dim)

def RDM_VPSDE(feat_dim):
    """Creates a Likelihood object for Variational Poisson SDE with Dependency Detection Model.

    Args:
        feat_dim (int): The feature dimension of the model.

    Returns:
        Likelihood: The Likelihood object.

    """
    return Likelihood(VPSDE(beta_min=0.5, beta_max=15, N=1000),feat_dim=feat_dim)



import optuna
import pathlib
import ood_detectors.ops_utils as ops_utils
from evaluate import run
import multiprocessing as mp
import random
import tqdm
# data = {
#         'features': features_vectors,
#         'encoder': encoder_name,
#         'target_dataset': dataset_name,
#         'dataset': name,
#         'type': type_name
#     }

def select_trial(trial,method):
    conf = {}
    if method == 'Residual':
        conf['dims'] = trial.suggest_float('Residual.dims', 0, 1)
    else:
        conf['n_epochs'] = trial.suggest_int('n_epochs', 100, 500, step=100)
        conf['bottleneck_channels'] = trial.suggest_int('bottleneck_channels', 256, 2048, step = 256)
        conf['num_res_blocks'] = trial.suggest_int('num_res_blocks', 3, 15)
        conf['time_embed_dim'] = trial.suggest_int('time_embed_dim', 256, 1024, step = 256)
        conf['dropout'] = trial.suggest_float('dropout', 0.0, 0.5)
        conf['lr'] = trial.suggest_float('lr', 1e-6, 1e-2, log=True)
        # conf['beta1'] = trial.suggest_float('beta1', 0.5, 0.999)
        # conf['beta2'] = trial.suggest_float('beta2', 0.9, 0.999)
        # conf['eps'] = trial.suggest_float('eps', 1e-12, 1e-6, log=True)
        # conf['weight_decay'] = trial.suggest_float('weight_decay', 0.0, 1e-3)
        if method != 'subVPSDE':
            conf['continuous'] = trial.suggest_categorical('continuous', [True, False])
        else:
            conf['continuous'] = True
        if conf['continuous']:
            conf['likelihood_weighting'] = trial.suggest_categorical('likelihood_weighting', [True, False])
        else:
            conf['likelihood_weighting'] = False
        conf['reduce_mean'] = trial.suggest_categorical('reduce_mean', [True, False])
        if method == 'VESDE':
            conf['sigma_min'] = trial.suggest_float('sigma_min', 0.01, 0.1)
            conf['sigma_max'] = trial.suggest_int('beta_max', 10, 60, step=5)
        elif method == 'VPSDE':
            conf['beta_min'] = trial.suggest_float('beta_min', 0.0, 1.0)
            conf['beta_max'] = trial.suggest_int('beta_max', 10, 30, step=5)
        elif method == 'subVPSDE':
            conf['beta_min'] = trial.suggest_float('beta_min', 0.0, 1.0)
            conf['beta_max'] = trial.suggest_int('beta_max', 10, 30, step=5)
        else:
            raise ValueError(f'Unknown method: {method}')
    return conf


def objective(trial, data, encoders, datasets, method, device, verbose=True):
    conf = select_trial(trial, method)
    ids = []
    faroods = []
    nearoods = []
    if verbose:
        bar = tqdm.tqdm(total=len(encoders)*len(datasets))
    random.shuffle(encoders)
    for encoder in encoders:
        random.shuffle(datasets)
        for dataset in datasets:
            if verbose:
                bar.set_description(f'Method: {method}, Encoder: {encoder}, Dataset: {dataset}')
            results = run(conf, data, encoder, dataset, method, device)
            auc = results['id']["AUC"]
            fpr = results['id']["FPR_95"]
            loss = results['id']['loss']
            score_id = results['id']['score_id']
            score_ref = results['id']['score_ref']
            farood = sum([v["AUC"] for v in results['farood'].values()]) / len(results['farood'])
            nearood = sum([v["AUC"] for v in results['nearood'].values()]) / len(results['nearood'])
            ids.append(auc)
            faroods.append(farood)
            nearoods.append(nearood)
            trial.set_user_attr(f'{encoder}_{dataset}_id_AUC', float(auc))
            trial.set_user_attr(f'{encoder}_{dataset}_id_FPR95', float(fpr))
            trial.set_user_attr(f'{encoder}_{dataset}_id_loss', float(loss))
            trial.set_user_attr(f'{encoder}_{dataset}_score_id', float(score_id))
            trial.set_user_attr(f'{encoder}_{dataset}_score_ref', float(score_ref))
            for d_name, v in results['farood'].items():
                for m, value in v.items():
                    trial.set_user_attr(f'{encoder}_{dataset}_farood_{d_name}_{m}', float(value))
            for d_name, v in results['nearood'].items():
                for m, value in v.items():
                    trial.set_user_attr(f'{encoder}_{dataset}_nearood_{d_name}_{m}', float(value))
                
            if verbose:
                bar.set_postfix(id=auc, farood=farood, nearood=nearood)
                bar.update()
    id = sum(ids) / len(ids)
    farood = sum(faroods) / len(faroods)
    nearood = sum(nearoods) / len(nearoods)
    return abs(id-0.5), farood, nearood

def ask_tell_optuna(objective_func, data, encoders, datasets, method, device):
    study_name = f'{method}'
    db = f'sqlite:///optuna_v2.db'
    study = optuna.create_study(directions=['minimize', 'maximize', 'maximize'], study_name=study_name, storage=db, load_if_exists=True)
    trial = study.ask()
    res = objective_func(trial, data, encoders, datasets, method, device)
    study.tell(trial, res)
        

def main():
    # features = pathlib.Path(r"H:\arty\data\features_opt")
    features = pathlib.Path("/mnt/data/arty/data/features_opt")
    features_data = {}
    all_pkl = list(features.rglob("*.pkl"))
    for path in all_pkl:
        parts = path.parts[len(features.parts):]
        tmp = features_data
        for p in parts[:-2]:
            if p not in tmp:
                tmp[p] = {}
            tmp = tmp[p]
        else:
            tmp[parts[-2]] = path
    # encoders = ['repvgg', 'resnet50d', 'swin', 'deit', 'dino', 'dinov2', 'vit', 'clip']
    encoders = ['repvgg', 'resnet50d', 'swin', 'deit', 'dino', 'dinov2', 'vit']
    #['resnet18_32x32_cifar10_open_ood', 'resnet18_32x32_cifar100_open_ood', 'resnet18_224x224_imagenet200_open_ood', 'resnet50_224x224_imagenet_open_ood']
    # datasets = ['imagenet', 'imagenet200', 'cifar10', 'cifar100', 'covid', 'mnist']
    datasets = ['imagenet']
    # methods = ['VESDE', 'VPSDE', 'subVPSDE', 'Residual']
    methods = ['VESDE', 'VPSDE',]
    jobs = []
    for m in methods:
        jobs.append((objective, features_data, encoders, datasets, m))
       
    trials = 100
    gpu_nodes = [0, 1, 2, 3] * 2
    random.shuffle(jobs)
    print(f'Running {len(jobs)} jobs...')
    ops_utils.parallelize(ask_tell_optuna, jobs*trials, gpu_nodes, verbose=True, timeout=60*60*24)

if __name__ == '__main__':
    mp.freeze_support()
    main()



from sklearn import metrics
import numpy as np


def auc(ind_conf, ood_conf):
    """
    Calculate the Area Under the ROC Curve (AUC) given the in-distribution and out-of-distribution confidence scores.

    Args:
        ind_conf (numpy.ndarray): Confidence scores for in-distribution samples.
        ood_conf (numpy.ndarray): Confidence scores for out-of-distribution samples.

    Returns:
        float: The AUC value.
    """
    conf = np.concatenate((ind_conf, ood_conf))
    ind_indicator = np.concatenate((np.ones_like(ind_conf), np.zeros_like(ood_conf)))

    fpr, tpr, _ = metrics.roc_curve(ind_indicator, conf)
    auroc = metrics.auc(fpr, tpr)

    return auroc


def num_fp_at_recall(ind_conf, ood_conf, tpr):
    """
    Calculate the number of false positives at a given recall rate.

    Args:
        ind_conf (numpy.ndarray): Confidence scores for in-distribution samples.
        ood_conf (numpy.ndarray): Confidence scores for out-of-distribution samples.
        tpr (float): Target recall rate.

    Returns:
        int: The number of false positives.
    """
    num_ind = len(ind_conf)

    if num_ind == 0 and len(ood_conf) == 0:
        return 0
    if num_ind == 0:
        return 0

    recall_num = int(np.floor(tpr * num_ind))
    thresh = np.sort(ind_conf)[-recall_num]
    num_fp = np.sum(ood_conf >= thresh)
    return num_fp


def fpr(ind_conf, ood_conf, tpr=0.95):
    """
    Calculate the False Positive Rate (FPR) at a given recall rate.

    Args:
        ind_conf (numpy.ndarray): Confidence scores for in-distribution samples.
        ood_conf (numpy.ndarray): Confidence scores for out-of-distribution samples.
        tpr (float, optional): Target recall rate. Defaults to 0.95.

    Returns:
        float: The False Positive Rate.
    """
    num_fp = num_fp_at_recall(ind_conf, ood_conf, tpr)
    num_ood = len(ood_conf)
    fpr = num_fp / max(1, num_ood)
    return fpr


def eval_ood(
    model,
    dataset,
    reference_dataset,
    ood_datasets,
    batch_size,
    num_workers=0,
    recall=0.95,
    verbose=True,
):
    """
    Evaluate the out-of-distribution (OOD) detection performance of a model.

    Args:
        model: The model to evaluate.
        dataset: The in-distribution dataset.
        reference_dataset: The reference dataset.
        ood_datasets: List of out-of-distribution datasets.
        batch_size (int): Batch size for prediction.
        num_workers (int, optional): Number of workers for data loading. Defaults to 0.
        recall (float, optional): Target recall rate for FPR calculation. Defaults to 0.95.
        verbose (bool, optional): Whether to print evaluation progress. Defaults to True.

    Returns:
        dict: A dictionary containing the evaluation results.
    """
    if verbose:
        print("Running eval. In-distribution data")
    score_id = model.predict(dataset, batch_size, num_workers, verbose=verbose)
    if verbose:
        print("Running eval. Reference data")
    score_ref = model.predict(
        reference_dataset, batch_size, num_workers, verbose=verbose
    )
    ref_auc = auc(-score_ref, -score_id)
    ref_fpr = fpr(-score_ref, -score_id, recall)
    if verbose:
        print(f"AUC: {ref_auc:.4f}, FPR: {ref_fpr:.4f}")
    score_oods = []
    auc_oods = []
    fpr_oods = []
    for i, ood_dataset in enumerate(ood_datasets):
        if verbose:
            print(f"Running eval. Out-of-distribution data {i+1}/{len(ood_datasets)}")
        score_ood = model.predict(ood_dataset, batch_size, num_workers, verbose=verbose)
        score_oods.append(score_ood)
        auc_ood = auc(-score_ref, -score_ood)
        auc_oods.append(auc_ood)
        fpr_ood = fpr(-score_ref, -score_ood, recall)
        fpr_oods.append(fpr_ood)
        if verbose:
            print(f"AUC: {auc_ood:.4f}, FPR: {fpr_ood:.4f}")

    return {
        "score": score_id,
        "score_ref": score_ref,
        "ref_auc": ref_auc,
        "ref_fpr": ref_fpr,
        "score_oods": score_oods,
        "auc": auc_oods,
        "fpr": fpr_oods,
    }



import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import seaborn as sns
import numpy as np
import pathlib


def plot(eval_data, id_name, ood_names, encoder, model, out_dir='figs', config=None, verbose=True, train_loss=None):
    if verbose:
        print('Generating plots...')
    # Unpack eval_data
    score, score_ref = eval_data['score'], eval_data['score_ref']
    ref_auc, ref_fpr = eval_data['ref_auc'], eval_data['ref_fpr']
    score_oods, auc_oods, fpr_oods = eval_data['score_oods'], eval_data['auc'], eval_data['fpr']

    # Create a figure with subplots
    fig, axs = plt.subplots(2, 2, figsize=(15, 10))  # Adjust the size as needed
    fig.suptitle(f'{model} Evaluation on {encoder}')

    def add_shadow(ax, data): 
        if data.var() > 1e-6:
            l = ax.lines[-1]
            x = l.get_xydata()[:,0]
            y = l.get_xydata()[:,1]
            ax.fill_between(x,y, alpha=0.1)
            # Calculate and plot the mean
            mean_value = np.mean(data)
            line_color = l.get_color()
            ax.axvline(mean_value, color=line_color, linestyle=':', linewidth=1.5)
    # Subplot 1: KDE plots
    sns.kdeplot(data=score, bw_adjust=.2, ax=axs[0, 0], label=f'{id_name} training: {np.mean(score):.2f}')
    add_shadow(axs[0, 0], score)

    sns.kdeplot(data=score_ref, bw_adjust=.2, ax=axs[0, 0], label=f'{id_name} validation: {np.mean(score_ref):.2f}')
    add_shadow(axs[0, 0], score_ref)

    for ood_name, score_ood in zip(ood_names, score_oods):
        sns.kdeplot(data=score_ood, bw_adjust=.2, ax=axs[0, 0], label=f'{ood_name}: {np.mean(score_ood):.2f}')
        add_shadow(axs[0, 0], score_ood)
    # axs[0, 0].xaxis.set_major_locator(ticker.MultipleLocator(base=0.5))
    # axs[0, 0].xaxis.set_minor_locator(ticker.MultipleLocator(base=0.1))
    axs[0, 0].set_title('Density Plots')
    axs[0, 0].set_xlabel('bits/dim')
    axs[0, 0].set_ylabel('Density')
    # axs[0, 0].set_xlim(6.5, 8)
    # axs[0, 0].grid(True)
    axs[0, 0].legend()

    # Subplot 2: Bar chart for AUC and FPR
    x = np.arange(len(ood_names)+1)  # the label locations
    width = 0.35  # the width of the bars
    disp_auc = [ref_auc] + auc_oods
    disp_fpr = [ref_fpr] + fpr_oods
    rects1 = axs[0, 1].bar(x - width/2, disp_auc, width, label='AUC', alpha=0.6)
    rects2 = axs[0, 1].bar(x + width/2, disp_fpr, width, label='FPR', alpha=0.6)
    axs[0, 1].set_ylabel('Metric Value')
    axs[0, 1].yaxis.set_major_locator(ticker.MultipleLocator(base=0.1))
    axs[0, 1].yaxis.set_minor_locator(ticker.MultipleLocator(base=0.05))
    axs[0, 1].set_title(f'AUC and FPR Metrics\nMean AUC: {np.mean(disp_auc[1:]):.2f}, Mean FPR: {np.mean(disp_fpr[1:]):.2f}')
    axs[0, 1].set_xticks(x)
    names = [f'{name}\nAUC: {auc:.2f}\nFPR: {fpr:.2f}' for name, auc, fpr in zip([id_name]+list(ood_names), disp_auc, disp_fpr)]
    axs[0, 1].set_xticklabels(names)
    axs[0, 1].grid(True)
    axs[0, 1].set_ylim([0, 1])
    axs[0, 1].legend()
    # add line at 0.5
    axs[0, 1].axhline(0.5, color='red', linestyle='--', linewidth=1.5) 

    if train_loss is not None:
        # Subplot 3: Training loss over time
        axs[1, 0].plot(train_loss, label='Training Loss')
        axs[1, 0].set_xlabel('Epochs')
        axs[1, 0].set_ylabel('Loss')
        axs[1, 0].set_title('Training Loss Over Time')
        axs[1, 0].grid(True)
        axs[1, 0].legend()
    else:
        axs[1, 0].axis('off')

    # Subplot 4: Configuration display
    if config is not None:
        config_text = "\n".join([f"{key}: {value}" for key, value in config.items()])
        axs[1, 1].text(0.5, 0.5, config_text, ha='center', va='center', fontsize=12, transform=axs[1, 1].transAxes)
        axs[1, 1].set_title('Configuration')
    axs[1, 1].axis('off')

    plt.tight_layout(rect=[0, 0.03, 1, 0.95])  # Adjust the layout

    # Save the figure
    out_dir = pathlib.Path(out_dir) / encoder / id_name
    out_dir.mkdir(exist_ok=True, parents=True)
    filename = f"{encoder}_{id_name}_{int(np.mean(disp_auc[1:])*100)}.svg"
    plt.savefig(out_dir / filename, bbox_inches='tight')
    if verbose:
        plt.show()


import time, random
import multiprocessing as mp

try:
    mp.set_start_method('spawn')
except RuntimeError:
    pass
class dummy:
    @staticmethod
    def is_alive():
        return False
    @staticmethod
    def close():
        pass
    @staticmethod
    def terminate():
        pass
    @staticmethod
    def join():
        pass
    @staticmethod
    def kill():
        pass

class Worker:
    def __init__(self, id, node):
        self.id = id
        self.node = node
        self.process = dummy
        self.start_time = time.perf_counter()


def parallelize(func, jobs, gpu_nodes, verbose=True, timeout=60*60*24):
    """
    Parallelize the execution of a function on multiple GPUs.

    Args:
        func (function): The function to be executed in parallel.
        jobs (list): A list of jobs to be executed by the function.
        gpu_nodes (list): A list of GPU nodes to be used for execution.
        verbose (bool, optional): Whether to print verbose output. Defaults to True.
        timeout (int, optional): Timeout in seconds for each job. Defaults to 3600.

    Returns:
        None
    """
    if verbose:
        print(f'Launching {len(jobs)} jobs on {len(set(gpu_nodes))} GPUs. {len(gpu_nodes)//len(set(gpu_nodes))} jobs per GPU in parallel..')
    workers = [Worker(id, node) for id, node in enumerate(gpu_nodes)]
    while len(jobs) > 0:
        random.shuffle(workers)
        for worker in workers:
            if time.perf_counter() - worker.start_time > timeout:
                if verbose:
                    print(f'Job on cuda:{worker.node} in slot {worker.id} timed out. Killing it...')
                worker.process.terminate()
            if not worker.process.is_alive():
                worker.process.kill()
                if verbose:
                    print(f'Launching job on cuda:{worker.node} in slot {worker.id}. {len(jobs)} jobs to left...')
                if len(jobs) == 0:
                    break
                args = list(jobs.pop())
                args.append(f'cuda:{worker.node}')
                p = mp.Process(target=func, args=args)
                p.start()
                worker.process = p
                worker.start_time = time.perf_counter()
                time.sleep(1)
        time.sleep(1)
    for worker in workers:
        worker.process.join()
    if verbose:
        print('Done!')


# Modified from https://raw.githubusercontent.com/fadel/pytorch_ema/master/torch_ema/ema.py

from __future__ import division
from __future__ import unicode_literals

import torch


# Partially based on: https://github.com/tensorflow/tensorflow/blob/r1.13/tensorflow/python/training/moving_averages.py
class ExponentialMovingAverage:
  """
  Maintains (exponential) moving average of a set of parameters.
  """

  def __init__(self, parameters, decay, use_num_updates=True):
    """
    Args:
      parameters: Iterable of `torch.nn.Parameter`; usually the result of
        `model.parameters()`.
      decay: The exponential decay.
      use_num_updates: Whether to use number of updates when computing
        averages.
    """
    if decay < 0.0 or decay > 1.0:
      raise ValueError('Decay must be between 0 and 1')
    self.decay = decay
    self.num_updates = 0 if use_num_updates else None
    self.shadow_params = [p.clone().detach()
                          for p in parameters if p.requires_grad]
    self.collected_params = []

  def update(self, parameters):
    """
    Update currently maintained parameters.

    Call this every time the parameters are updated, such as the result of
    the `optimizer.step()` call.

    Args:
      parameters: Iterable of `torch.nn.Parameter`; usually the same set of
        parameters used to initialize this object.
    """
    decay = self.decay
    if self.num_updates is not None:
      self.num_updates += 1
      decay = min(decay, (1 + self.num_updates) / (10 + self.num_updates))
    one_minus_decay = 1.0 - decay
    with torch.no_grad():
      parameters = [p for p in parameters if p.requires_grad]
      for s_param, param in zip(self.shadow_params, parameters):
        s_param.sub_(one_minus_decay * (s_param - param))

  def copy_to(self, parameters):
    """
    Copy current parameters into given collection of parameters.

    Args:
      parameters: Iterable of `torch.nn.Parameter`; the parameters to be
        updated with the stored moving averages.
    """
    parameters = [p for p in parameters if p.requires_grad]
    for s_param, param in zip(self.shadow_params, parameters):
      if param.requires_grad:
        param.data.copy_(s_param.data)

  def store(self, parameters):
    """
    Save the current parameters for restoring later.

    Args:
      parameters: Iterable of `torch.nn.Parameter`; the parameters to be
        temporarily stored.
    """
    self.collected_params = [param.clone() for param in parameters]

  def restore(self, parameters):
    """
    Restore the parameters stored with the `store` method.
    Useful to validate the model with EMA parameters without affecting the
    original optimization process. Store the parameters before the
    `copy_to` method. After validation (or model saving), use this to
    restore the former parameters.

    Args:
      parameters: Iterable of `torch.nn.Parameter`; the parameters to be
        updated with the stored parameters.
    """
    for c_param, param in zip(self.collected_params, parameters):
      param.data.copy_(c_param.data)

  def state_dict(self):
    return dict(decay=self.decay, num_updates=self.num_updates,
                shadow_params=self.shadow_params)

  def load_state_dict(self, state_dict):
    self.decay = state_dict['decay']
    self.num_updates = state_dict['num_updates']
    self.shadow_params = state_dict['shadow_params']